from .generator import *  # noqa
from .adasyn import *  # noqa
from .smote import *  # noqa
from .randomoversampler import *  # noqa
from .tvae import *  # noqa
from .ctgan import *  # noqa
from .gaussiancopula import *  # noqa
from .copulagan import *  # noqa
from .ctabgan import *  # noqa
from .ctabganplus import *  # noqa
from .autodiffusion import *  # noqa
from .forestdiffusion import *  # noqa
from .tabula import *  # noqua
from .great import *  # noqua



import numpy as np
import pandas as pd
import torch
import torch.utils.data
import torch.optim as optim
from torch.optim import Adam
from torch.nn import functional as F
from torch.nn import (
    Dropout,
    LeakyReLU,
    Linear,
    Module,
    ReLU,
    Sequential,
    Conv2d,
    ConvTranspose2d,
    BatchNorm2d,
    Sigmoid,
    init,
    BCELoss,
    CrossEntropyLoss,
    SmoothL1Loss,
)
from .transformer import ImageTransformer, DataTransformer


def random_choice_prob_index_sampling(probs, col_idx):
    """
    Used to sample a specific category within a chosen one-hot-encoding representation

    Inputs:
    1) probs -> probability mass distribution of categories
    2) col_idx -> index used to identify any given one-hot-encoding

    Outputs:
    1) option_list -> list of chosen categories

    """

    option_list = []
    for i in col_idx:
        # for improved stability
        pp = probs[i] + 1e-6
        pp = pp / sum(pp)
        # sampled based on given probability mass distribution of categories within the given one-hot-encoding
        option_list.append(np.random.choice(np.arange(len(probs[i])), p=pp))

    return np.array(option_list).reshape(col_idx.shape)


class Condvec(object):
    """
    This class is responsible for sampling conditional vectors to be supplied to the generator

    Variables:
    1) model -> list containing an index of highlighted categories in their corresponding one-hot-encoded represenations
    2) interval -> an array holding the respective one-hot-encoding starting positions and sizes
    3) n_col -> total no. of one-hot-encoding representations
    4) n_opt -> total no. of distinct categories across all one-hot-encoding representations
    5) p_log_sampling -> list containing log of probability mass distribution of categories within their respective one-hot-encoding representations
    6) p_sampling -> list containing probability mass distribution of categories within their respective one-hot-encoding representations

    Methods:
    1) __init__() -> takes transformed input data with respective column information to compute class variables
    2) sample_train() -> used to sample the conditional vector during training of the model
    3) sample() -> used to sample the conditional vector for generating data after training is finished

    """

    def __init__(self, data, output_info):

        self.model = []
        self.interval = []
        self.n_col = 0
        self.n_opt = 0
        self.p_log_sampling = []
        self.p_sampling = []

        # iterating through the transformed input data columns
        st = 0
        for item in output_info:
            # ignoring columns that do not represent one-hot-encodings
            if item[1] == "tanh":
                st += item[0]
                continue
            elif item[1] == "softmax":
                # using starting (st) and ending (ed) position of any given one-hot-encoded representation to obtain relevant information
                ed = st + item[0]
                self.model.append(np.argmax(data[:, st:ed], axis=-1))
                self.interval.append((self.n_opt, item[0]))
                self.n_col += 1
                self.n_opt += item[0]
                freq = np.sum(data[:, st:ed], axis=0)
                log_freq = np.log(freq + 1)
                log_pmf = log_freq / np.sum(log_freq)
                self.p_log_sampling.append(log_pmf)
                pmf = freq / np.sum(freq)
                self.p_sampling.append(pmf)
                st = ed

        self.interval = np.asarray(self.interval)

    def sample_train(self, batch):
        """
        Used to create the conditional vectors for feeding it to the generator during training

        Inputs:
        1) batch -> no. of data records to be generated in a batch

        Outputs:
        1) vec -> a matrix containing a conditional vector for each data point to be generated
        2) mask -> a matrix to identify chosen one-hot-encodings across the batch
        3) idx -> list of chosen one-hot encoding across the batch
        4) opt1prime -> selected categories within chosen one-hot-encodings

        """

        if self.n_col == 0:
            return None
        batch = batch

        # each conditional vector in vec is a one-hot vector used to highlight a specific category across all possible one-hot-encoded representations
        # (i.e., including modes of continuous and mixed columns)
        vec = np.zeros((batch, self.n_opt), dtype="float32")

        # choosing one specific one-hot-encoding from all possible one-hot-encoded representations
        idx = np.random.choice(np.arange(self.n_col), batch)

        # matrix of shape (batch x total no. of one-hot-encoded representations) with 1 in indexes of chosen representations and 0 elsewhere
        mask = np.zeros((batch, self.n_col), dtype="float32")
        mask[np.arange(batch), idx] = 1

        # producing a list of selected categories within each of selected one-hot-encoding representation
        opt1prime = random_choice_prob_index_sampling(self.p_log_sampling, idx)

        # assigning the appropriately chosen category for each corresponding conditional vector
        for i in np.arange(batch):
            vec[i, self.interval[idx[i], 0] + opt1prime[i]] = 1

        return vec, mask, idx, opt1prime

    def sample(self, batch):
        """
        Used to create the conditional vectors for feeding it to the generator after training is finished

        Inputs:
        1) batch -> no. of data records to be generated in a batch

        Outputs:
        1) vec -> an array containing a conditional vector for each data point to be generated
        """

        if self.n_col == 0:
            return None

        batch = batch

        # each conditional vector in vec is a one-hot vector used to highlight a specific category across all possible one-hot-encoded representations
        # (i.e., including modes of continuous and mixed columns)
        vec = np.zeros((batch, self.n_opt), dtype="float32")

        # choosing one specific one-hot-encoding from all possible one-hot-encoded representations
        idx = np.random.choice(np.arange(self.n_col), batch)

        # producing a list of selected categories within each of selected one-hot-encoding representation
        opt1prime = random_choice_prob_index_sampling(self.p_sampling, idx)

        # assigning the appropriately chosen category for each corresponding conditional vector
        for i in np.arange(batch):
            vec[i, self.interval[idx[i], 0] + opt1prime[i]] = 1

        return vec


def cond_loss(data, output_info, c, m):
    """
    Used to compute the conditional loss for ensuring the generator produces the desired category as specified by the conditional vector

    Inputs:
    1) data -> raw data synthesized by the generator
    2) output_info -> column informtion corresponding to the data transformer
    3) c -> conditional vectors used to synthesize a batch of data
    4) m -> a matrix to identify chosen one-hot-encodings across the batch

    Outputs:
    1) loss -> conditional loss corresponding to the generated batch

    """

    # used to store cross entropy loss between conditional vector and all generated one-hot-encodings
    tmp_loss = []
    # counter to iterate generated data columns
    st = 0
    # counter to iterate conditional vector
    st_c = 0
    # iterating through column information
    for item in output_info:
        # ignoring numeric columns
        if item[1] == "tanh":
            st += item[0]
            continue
        # computing cross entropy loss between generated one-hot-encoding and corresponding encoding of conditional vector
        elif item[1] == "softmax":
            ed = st + item[0]
            ed_c = st_c + item[0]
            tmp = F.cross_entropy(
                data[:, st:ed], torch.argmax(c[:, st_c:ed_c], dim=1), reduction="none"
            )
            tmp_loss.append(tmp)
            st = ed
            st_c = ed_c

    # computing the loss across the batch only and only for the relevant one-hot-encodings by applying the mask
    tmp_loss = torch.stack(tmp_loss, dim=1)
    loss = (tmp_loss * m).sum() / data.size()[0]

    return loss


class Sampler(object):
    """
    This class is used to sample the transformed real data according to the conditional vector

    Variables:
    1) data -> real transformed input data
    2) model -> stores the index values of data records corresponding to any given selected categories for all columns
    3) n -> size of the input data

    Methods:
    1) __init__() -> initiates the sampler object and stores class variables
    2) sample() -> takes as input the number of rows to be sampled (n), chosen column (col)
                   and category within the column (opt) to sample real records accordingly
    """

    def __init__(self, data, output_info):

        super(Sampler, self).__init__()

        self.data = data
        self.model = []
        self.n = len(data)

        # counter to iterate through columns
        st = 0
        # iterating through column information
        for item in output_info:
            # ignoring numeric columns
            if item[1] == "tanh":
                st += item[0]
                continue
            # storing indices of data records for all categories within one-hot-encoded representations
            elif item[1] == "softmax":
                ed = st + item[0]
                tmp = []
                # iterating through each category within a one-hot-encoding
                for j in range(item[0]):
                    # storing the relevant indices of data records for the given categories
                    tmp.append(np.nonzero(data[:, st + j])[0])
                self.model.append(tmp)
                st = ed

    def sample(self, n, col, opt):

        # if there are no one-hot-encoded representations, we may ignore sampling using a conditional vector
        if col is None:
            idx = np.random.choice(np.arange(self.n), n)
            return self.data[idx]

        # used to store relevant indices of data records based on selected category within a chosen one-hot-encoding
        idx = []

        # sampling a data record index randomly from all possible indices that meet the given criteria of the chosen category and one-hot-encoding
        for c, o in zip(col, opt):
            idx.append(np.random.choice(self.model[c][o]))

        return self.data[idx]


def get_st_ed(target_col_index, output_info):
    """
    Used to obtain the start and ending positions of the target column as per the transformed data to be used by the classifier

    Inputs:
    1) target_col_index -> column index of the target column used for machine learning tasks (binary/multi-classification) in the raw data
    2) output_info -> column information corresponding to the data after applying the data transformer

    Outputs:
    1) starting (st) and ending (ed) positions of the target column as per the transformed data

    """
    # counter to iterate through columns
    st = 0
    # counter to check if the target column index has been reached
    c = 0
    # counter to iterate through column information
    tc = 0
    # iterating until target index has reached to obtain starting position of the one-hot-encoding used to represent target column in transformed data
    for item in output_info:
        # exiting loop if target index has reached
        if c == target_col_index:
            break
        if item[1] == "tanh":
            st += item[0]
        elif item[1] == "softmax":
            st += item[0]
            c += 1
        tc += 1

    # obtaining the ending position by using the dimension size of the one-hot-encoding used to represent the target column
    ed = st + output_info[tc][0]

    return (st, ed)


class Classifier(Module):
    """
    This class represents the classifier module used along side the discriminator to train the generator network

    Variables:
    1) dim -> column dimensionality of the transformed input data after removing target column
    2) class_dims -> list of dimensions used for the hidden layers of the classifier network
    3) str_end -> tuple containing the starting and ending positions of the target column in the transformed input data

    Methods:
    1) __init__() -> initializes and builds the layers of the classifier module
    2) forward() -> executes the forward pass of the classifier module on the corresponding input data and
                    outputs the predictions and corresponding true labels for the target column

    """

    def __init__(self, input_dim, class_dims, st_ed):
        super(Classifier, self).__init__()
        # subtracting the target column size from the input dimensionality
        self.dim = input_dim - (st_ed[1] - st_ed[0])
        # storing the starting and ending positons of the target column in the input data
        self.str_end = st_ed

        # building the layers of the network with same hidden layers as discriminator
        seq = []
        tmp_dim = self.dim
        for item in list(class_dims):
            seq += [Linear(tmp_dim, item), LeakyReLU(0.2), Dropout(0.5)]
            tmp_dim = item

        # in case of binary classification the last layer outputs a single numeric value which is squashed to a probability with sigmoid
        if (st_ed[1] - st_ed[0]) == 2:
            seq += [Linear(tmp_dim, 1), Sigmoid()]
        # in case of multi-classs classification, the last layer outputs an array of numeric values associated to each class
        else:
            seq += [Linear(tmp_dim, (st_ed[1] - st_ed[0]))]

        self.seq = Sequential(*seq)

    def forward(self, input):

        # true labels obtained from the input data
        label = torch.argmax(input[:, self.str_end[0] : self.str_end[1]], axis=-1)

        # input to be fed to the classifier module
        new_imp = torch.cat(
            (input[:, : self.str_end[0]], input[:, self.str_end[1] :]), 1
        )

        # returning predictions and true labels for binary/multi-class classification
        if (self.str_end[1] - self.str_end[0]) == 2:
            return self.seq(new_imp).view(-1), label
        else:
            return self.seq(new_imp), label


class Discriminator(Module):
    """
    This class represents the discriminator network of the model

    Variables:
    1) seq -> layers of the network used for making the final prediction of the discriminator model
    2) seq_info -> layers of the discriminator network used for computing the information loss

    Methods:
    1) __init__() -> initializes and builds the layers of the discriminator model
    2) forward() -> executes a forward pass on the input data to output the final predictions and corresponding
                    feature information associated with the penultimate layer used to compute the information loss

    """

    def __init__(self, layers):
        super(Discriminator, self).__init__()
        self.seq = Sequential(*layers)
        self.seq_info = Sequential(*layers[: len(layers) - 2])

    def forward(self, input):
        return (self.seq(input)), self.seq_info(input)


class Generator(Module):
    """
    This class represents the discriminator network of the model

    Variables:
    1) seq -> layers of the network used by the generator

    Methods:
    1) __init__() -> initializes and builds the layers of the generator model
    2) forward() -> executes a forward pass using noise as input to generate data

    """

    def __init__(self, layers):
        super(Generator, self).__init__()
        self.seq = Sequential(*layers)

    def forward(self, input):
        return self.seq(input)


def determine_layers_disc(side, num_channels):
    """
    This function describes the layers of the discriminator network as per DCGAN (https://pytorch.org/tutorials/beginner/dcgan_faces_tutorial.html)

    Inputs:
    1) side -> height/width of the input fed to the discriminator
    2) num_channels -> no. of channels used to decide the size of respective hidden layers

    Outputs:
    1) layers_D -> layers of the discriminator network

    """

    # computing the dimensionality of hidden layers
    layer_dims = [(1, side), (num_channels, side // 2)]

    while layer_dims[-1][1] > 3 and len(layer_dims) < 4:
        # the number of channels increases by a factor of 2 whereas the height/width decreases by the same factor with each layer
        layer_dims.append((layer_dims[-1][0] * 2, layer_dims[-1][1] // 2))

    # constructing the layers of the discriminator network based on the recommendations mentioned in https://arxiv.org/abs/1511.06434
    layers_D = []
    for prev, curr in zip(layer_dims, layer_dims[1:]):
        layers_D += [
            Conv2d(prev[0], curr[0], 4, 2, 1, bias=False),
            BatchNorm2d(curr[0]),
            LeakyReLU(0.2, inplace=True),
        ]
    # last layer reduces the output to a single numeric value which is squashed to a probabability using sigmoid function
    layers_D += [Conv2d(layer_dims[-1][0], 1, layer_dims[-1][1], 1, 0), Sigmoid()]

    return layers_D


def determine_layers_gen(side, random_dim, num_channels):
    """
    This function describes the layers of the generator network

    Inputs:
    1) random_dim -> height/width of the noise matrix to be fed for generation
    2) num_channels -> no. of channels used to decide the size of respective hidden layers

    Outputs:
    1) layers_G -> layers of the generator network

    """

    # computing the dimensionality of hidden layers
    layer_dims = [(1, side), (num_channels, side // 2)]

    while layer_dims[-1][1] > 3 and len(layer_dims) < 4:
        layer_dims.append((layer_dims[-1][0] * 2, layer_dims[-1][1] // 2))

    # similarly constructing the layers of the generator network based on the recommendations mentioned in https://arxiv.org/abs/1511.06434
    # first layer of the generator takes the channel dimension of the noise matrix to the desired maximum channel size of the generator's layers
    layers_G = [
        ConvTranspose2d(
            random_dim,
            layer_dims[-1][0],
            layer_dims[-1][1],
            1,
            0,
            output_padding=0,
            bias=False,
        )
    ]

    # the following layers are then reversed with respect to the discriminator
    # such as the no. of channels reduce by a factor of 2 and height/width of generated image increases by the same factor with each layer
    for prev, curr in zip(reversed(layer_dims), reversed(layer_dims[:-1])):
        layers_G += [
            BatchNorm2d(prev[0]),
            ReLU(True),
            ConvTranspose2d(prev[0], curr[0], 4, 2, 1, output_padding=0, bias=True),
        ]

    return layers_G


def apply_activate(data, output_info):
    """
    This function applies the final activation corresponding to the column information associated with transformer

    Inputs:
    1) data -> input data generated by the model in the same format as the transformed input data
    2) output_info -> column information associated with the transformed input data

    Outputs:
    1) act_data -> resulting data after applying the respective activations

    """

    data_t = []
    # used to iterate through columns
    st = 0
    # used to iterate through column information
    for item in output_info:
        # for numeric columns a final tanh activation is applied
        if item[1] == "tanh":
            ed = st + item[0]
            data_t.append(torch.tanh(data[:, st:ed]))
            st = ed
        # for one-hot-encoded columns, a final gumbel softmax (https://arxiv.org/pdf/1611.01144.pdf) is used
        # to sample discrete categories while still allowing for back propagation
        elif item[1] == "softmax":
            ed = st + item[0]
            # note that as tau approaches 0, a completely discrete one-hot-vector is obtained
            data_t.append(F.gumbel_softmax(data[:, st:ed], tau=0.2))
            st = ed

    act_data = torch.cat(data_t, dim=1)

    return act_data


def weights_init(model):
    """
    This function initializes the learnable parameters of the convolutional and batch norm layers

    Inputs:
    1) model->  network for which the parameters need to be initialized

    Outputs:
    1) network with corresponding weights initialized using the normal distribution

    """

    classname = model.__class__.__name__

    if classname.find("Conv") != -1:
        init.normal_(model.weight.data, 0.0, 0.02)

    elif classname.find("BatchNorm") != -1:
        init.normal_(model.weight.data, 1.0, 0.02)
        init.constant_(model.bias.data, 0)


class CTABGANSynthesizer:
    """
    This class represents the main model used for training the model and generating synthetic data


    Variables:
    1) random_dim -> size of the noise vector fed to the generator
    2) class_dim -> tuple containing dimensionality of hidden layers for the classifier network
    3) num_channels -> no. of channels for deciding respective hidden layers of discriminator and generator networks
    4) dside -> height/width of the input data fed to discriminator network
    5) gside -> height/width of the input data generated by the generator network
    6) l2scale -> parameter to decide strength of regularization of the network based on constraining l2 norm of weights
    7) batch_size -> no. of records to be processed in each mini-batch of training
    8) epochs -> no. of epochs to train the model
    9) device -> type of device to be used for training (i.e., gpu/cpu)
    10) generator -> generator network from which data can be generated after training the model

    Methods:
    1) __init__() -> initializes the model with user specified parameters
    2) fit() -> takes the pre-processed training data and associated parameters as input to fit the CTABGANSynthesizer model
    3) sample() -> takes as input the no. of data rows to be generated and synthesizes the corresponding no. of data rows

    """

    def __init__(
        self,
        class_dim=(256, 256, 256, 256),
        random_dim=100,
        num_channels=64,
        l2scale=1e-5,
        batch_size=500,
        epochs=1,
    ):

        self.random_dim = random_dim
        self.class_dim = class_dim
        self.num_channels = num_channels
        self.dside = None
        self.gside = None
        self.l2scale = l2scale
        self.batch_size = batch_size
        self.epochs = epochs
        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        self.generator = None

    def fit(
        self,
        train_data=pd.DataFrame,
        categorical=[],
        mixed={},
        type={},
        progress=None,
        task=None,
    ):

        # obtaining the column index of the target column used for ML tasks
        problem_type = None
        target_index = None

        if type:
            problem_type = list(type.keys())[0]
            if problem_type:
                target_index = train_data.columns.get_loc(type[problem_type])

        # transforming pre-processed training data according to different data types
        # i.e., mode specific normalisation for numeric and mixed columns and one-hot-encoding for categorical columns
        self.transformer = DataTransformer(
            train_data=train_data, categorical_list=categorical, mixed_dict=mixed
        )
        self.transformer.fit()
        train_data = self.transformer.transform(train_data.values)
        # storing column size of the transformed training data
        data_dim = self.transformer.output_dim

        # initializing the sampler object to execute training-by-sampling
        data_sampler = Sampler(train_data, self.transformer.output_info)
        # initializing the condvec object to sample conditional vectors during training
        self.cond_generator = Condvec(train_data, self.transformer.output_info)

        # obtaining the desired height/width for converting tabular data records to square images for feeding it to discriminator network
        sides = [4, 8, 16, 24, 32, 64, 128]
        # the discriminator takes the transformed training data concatenated by the corresponding conditional vectors as input
        col_size_d = data_dim + self.cond_generator.n_opt
        for i in sides:
            if i * i >= col_size_d:
                self.dside = i
                break

        # obtaining the desired height/width for generating square images from the generator network that can be converted back to tabular domain
        sides = [4, 8, 16, 24, 32, 64, 128]
        col_size_g = data_dim
        for i in sides:
            if i * i >= col_size_g:
                self.gside = i
                break

        # constructing the generator and discriminator networks
        layers_G = determine_layers_gen(
            self.gside, self.random_dim + self.cond_generator.n_opt, self.num_channels
        )
        layers_D = determine_layers_disc(self.dside, self.num_channels)
        self.generator = Generator(layers_G).to(self.device)
        discriminator = Discriminator(layers_D).to(self.device)

        # assigning the respective optimizers for the generator and discriminator networks
        optimizer_params = dict(
            lr=2e-4, betas=(0.5, 0.9), eps=1e-3, weight_decay=self.l2scale
        )
        optimizerG = Adam(self.generator.parameters(), **optimizer_params)
        optimizerD = Adam(discriminator.parameters(), **optimizer_params)

        st_ed = None
        classifier = None
        optimizerC = None
        if target_index != None:
            # obtaining the one-hot-encoding starting and ending positions of the target column in the transformed data
            st_ed = get_st_ed(target_index, self.transformer.output_info)
            # configuring the classifier network and it's optimizer accordingly
            classifier = Classifier(data_dim, self.class_dim, st_ed).to(self.device)
            optimizerC = optim.Adam(classifier.parameters(), **optimizer_params)

        # initializing learnable parameters of the discrimnator and generator networks
        self.generator.apply(weights_init)
        discriminator.apply(weights_init)

        # initializing the image transformer objects for the generator and discriminator networks for transitioning between image and tabular domain
        self.Gtransformer = ImageTransformer(self.gside)
        self.Dtransformer = ImageTransformer(self.dside)

        # initiating the training by computing the number of iterations per epoch
        steps_per_epoch = max(1, len(train_data) // self.batch_size)
        for i in range(self.epochs):
            for _ in range(steps_per_epoch):

                # sampling noise vectors using a standard normal distribution
                noisez = torch.randn(
                    self.batch_size, self.random_dim, device=self.device
                )
                # sampling conditional vectors
                condvec = self.cond_generator.sample_train(self.batch_size)
                c, m, col, opt = condvec
                c = torch.from_numpy(c).to(self.device)
                m = torch.from_numpy(m).to(self.device)
                # concatenating conditional vectors and converting resulting noise vectors into the image domain to be fed to the generator as input
                noisez = torch.cat([noisez, c], dim=1)
                noisez = noisez.view(
                    self.batch_size, self.random_dim + self.cond_generator.n_opt, 1, 1
                )

                # sampling real data according to the conditional vectors and shuffling it before feeding to discriminator to isolate conditional loss on generator
                perm = np.arange(self.batch_size)
                np.random.shuffle(perm)
                real = data_sampler.sample(self.batch_size, col[perm], opt[perm])
                real = torch.from_numpy(real.astype("float32")).to(self.device)

                # storing shuffled ordering of the conditional vectors
                c_perm = c[perm]
                # generating synthetic data as an image
                fake = self.generator(noisez)
                # converting it into the tabular domain as per format of the trasformed training data
                faket = self.Gtransformer.inverse_transform(fake)
                # applying final activation on the generated data (i.e., tanh for numeric and gumbel-softmax for categorical)
                fakeact = apply_activate(faket, self.transformer.output_info)

                # the generated data is then concatenated with the corresponding condition vectors
                fake_cat = torch.cat([fakeact, c], dim=1)
                # the real data is also similarly concatenated with corresponding conditional vectors
                real_cat = torch.cat([real, c_perm], dim=1)

                # transforming the real and synthetic data into the image domain for feeding it to the discriminator
                real_cat_d = self.Dtransformer.transform(real_cat)
                fake_cat_d = self.Dtransformer.transform(fake_cat)

                # executing the gradient update step for the discriminator
                optimizerD.zero_grad()
                # computing the probability of the discriminator to correctly classify real samples hence y_real should ideally be close to 1
                y_real, _ = discriminator(real_cat_d)
                # computing the probability of the discriminator to correctly classify fake samples hence y_fake should ideally be close to 0
                y_fake, _ = discriminator(fake_cat_d)
                # computing the loss to essentially maximize the log likelihood of correctly classifiying real and fake samples as log(D(x))+log(1−D(G(z)))
                # or equivalently minimizing the negative of log(D(x))+log(1−D(G(z))) as done below
                loss_d = -(torch.log(y_real + 1e-4).mean()) - (
                    torch.log(1.0 - y_fake + 1e-4).mean()
                )
                # accumulating gradients based on the loss
                loss_d.backward()
                # computing the backward step to update weights of the discriminator
                optimizerD.step()

                # similarly sample noise vectors and conditional vectors
                noisez = torch.randn(
                    self.batch_size, self.random_dim, device=self.device
                )
                condvec = self.cond_generator.sample_train(self.batch_size)
                c, m, col, opt = condvec
                c = torch.from_numpy(c).to(self.device)
                m = torch.from_numpy(m).to(self.device)
                noisez = torch.cat([noisez, c], dim=1)
                noisez = noisez.view(
                    self.batch_size, self.random_dim + self.cond_generator.n_opt, 1, 1
                )

                # executing the gradient update step for the generator
                optimizerG.zero_grad()

                # similarly generating synthetic data and applying final activation
                fake = self.generator(noisez)
                faket = self.Gtransformer.inverse_transform(fake)
                fakeact = apply_activate(faket, self.transformer.output_info)
                # concatenating conditional vectors and converting it to the image domain to be fed to the discriminator
                fake_cat = torch.cat([fakeact, c], dim=1)
                fake_cat = self.Dtransformer.transform(fake_cat)

                # computing the probability of the discriminator classifiying fake samples as real
                # along with feature representaions of fake data resulting from the penultimate layer
                y_fake, info_fake = discriminator(fake_cat)
                # extracting feature representation of real data from the penultimate layer of the discriminator
                _, info_real = discriminator(real_cat_d)
                # computing the conditional loss to ensure the generator generates data records with the chosen category as per the conditional vector
                cross_entropy = cond_loss(faket, self.transformer.output_info, c, m)

                # computing the loss to train the generator where we want y_fake to be close to 1 to fool the discriminator
                # and cross_entropy to be close to 0 to ensure generator's output matches the conditional vector
                g = -(torch.log(y_fake + 1e-4).mean()) + cross_entropy
                # in order to backprop the gradient of separate losses w.r.t to the learnable weight of the network independently
                # we may use retain_graph=True in backward() method in the first back-propagated loss
                # to maintain the computation graph to execute the second backward pass efficiently
                g.backward(retain_graph=True)
                # computing the information loss by comparing means and stds of real/fake feature representations extracted from discriminator's penultimate layer
                loss_mean = torch.norm(
                    torch.mean(info_fake.view(self.batch_size, -1), dim=0)
                    - torch.mean(info_real.view(self.batch_size, -1), dim=0),
                    1,
                )
                loss_std = torch.norm(
                    torch.std(info_fake.view(self.batch_size, -1), dim=0)
                    - torch.std(info_real.view(self.batch_size, -1), dim=0),
                    1,
                )
                loss_info = loss_mean + loss_std
                # computing the finally accumulated gradients
                loss_info.backward()
                # executing the backward step to update the weights
                optimizerG.step()

                # the classifier module is used in case there is a target column associated with ML tasks
                if problem_type:

                    c_loss = None
                    # in case of binary classification, the binary cross entropy loss is used
                    if (st_ed[1] - st_ed[0]) == 2:
                        c_loss = BCELoss()
                    # in case of multi-class classification, the standard cross entropy loss is used
                    else:
                        c_loss = CrossEntropyLoss()

                    # updating the weights of the classifier
                    optimizerC.zero_grad()
                    # computing classifier's target column predictions on the real data along with returning corresponding true labels
                    real_pre, real_label = classifier(real)
                    if (st_ed[1] - st_ed[0]) == 2:
                        real_label = real_label.type_as(real_pre)
                    # computing the loss to train the classifier so that it can perform well on the real data
                    loss_cc = c_loss(real_pre, real_label)
                    loss_cc.backward()
                    optimizerC.step()

                    # updating the weights of the generator
                    optimizerG.zero_grad()
                    # generate synthetic data and apply the final activation
                    fake = self.generator(noisez)
                    faket = self.Gtransformer.inverse_transform(fake)
                    fakeact = apply_activate(faket, self.transformer.output_info)
                    # computing classifier's target column predictions on the fake data along with returning corresponding true labels
                    fake_pre, fake_label = classifier(fakeact)
                    if (st_ed[1] - st_ed[0]) == 2:
                        fake_label = fake_label.type_as(fake_pre)
                    # computing the loss to train the generator to improve semantic integrity between target column and rest of the data
                    loss_cg = c_loss(fake_pre, fake_label)
                    loss_cg.backward()
                    optimizerG.step()

            progress.update(task, completed=i)

    def sample(self, n):

        # turning the generator into inference mode to effectively use running statistics in batch norm layers
        self.generator.eval()
        # column information associated with the transformer fit to the pre-processed training data
        output_info = self.transformer.output_info

        # generating synthetic data in batches accordingly to the total no. required
        steps = n // self.batch_size + 1
        data = []
        for _ in range(steps):
            # generating synthetic data using sampled noise and conditional vectors
            noisez = torch.randn(self.batch_size, self.random_dim, device=self.device)
            condvec = self.cond_generator.sample(self.batch_size)
            c = condvec
            c = torch.from_numpy(c).to(self.device)
            noisez = torch.cat([noisez, c], dim=1)
            noisez = noisez.view(
                self.batch_size, self.random_dim + self.cond_generator.n_opt, 1, 1
            )
            fake = self.generator(noisez)
            faket = self.Gtransformer.inverse_transform(fake)
            fakeact = apply_activate(faket, output_info)
            data.append(fakeact.detach().cpu().numpy())

        data = np.concatenate(data, axis=0)

        # applying the inverse transform and returning synthetic data in a similar form as the original pre-processed training data
        result = self.transformer.inverse_transform(data)

        return result[0:n]



from . import Tuner
from gentab.generators import AutoDiffusion
from gentab.evaluators import Evaluator
from gentab.utils import console, SPINNER, REFRESH

import optuna


class AutoDiffusionTuner(Tuner):
    def __init__(
        self,
        evaluator: Evaluator,
        trials: int,
        *args,
        min_epochs: int = 500,
        max_epochs: int = 10000,
        min_batch: int = 64,
        max_batch: int = 8192,
        max_tries_per_batch: int = 8192,
        timeout: int = None,
        **kwargs,
    ) -> None:
        super().__init__(
            evaluator,
            trials,
            min_epochs=min_epochs,
            max_epochs=max_epochs,
            min_batch=min_batch,
            max_batch=max_batch,
            max_tries_per_batch=max_tries_per_batch,
            timeout=timeout,
        )

    def objective(self, trial: optuna.trial.Trial) -> float:
        n_epochs = trial.suggest_int("n_epochs", self.min_epochs, self.max_epochs)
        diff_n_epochs = trial.suggest_int(
            "diff_n_epochs", self.min_epochs, self.max_epochs
        )
        batch_size = trial.suggest_categorical("batch_size", self.batch_sizes)
        threshold = trial.suggest_float("threshold", 0.005, 0.02)
        weight_decay = trial.suggest_float("weight_decay", 1e-7, 1e-5, log=True)
        lr = trial.suggest_float("lr", 2e-5, 2e-3, log=True)
        hidden_size = trial.suggest_categorical("hidden_size", [64, 128, 256, 512])
        num_layers = trial.suggest_categorical("num_layers", [2, 4, 8])
        hidden_dims = trial.suggest_categorical(
            "hidden_dims",
            [
                (128, 256, 512, 256, 128),
                (256, 512, 1024, 512, 256),
                # (512, 1024, 2048, 1024, 512),
            ],
        )
        sigma = trial.suggest_int("sigma", 10, 40)
        T = trial.suggest_int("T", 50, 200, step=2)

        self.generator = AutoDiffusion(
            self.dataset,
            n_epochs=n_epochs,
            diff_n_epochs=diff_n_epochs,
            threshold=threshold,
            weight_decay=weight_decay,
            lr=lr,
            hidden_size=hidden_size,
            num_layers=num_layers,
            hidden_dims=hidden_dims,
            sigma=sigma,
            T=T,
            batch_size=batch_size,
            max_tries_per_batch=self.max_tries_per_batch,
        )
        self.generator.generate()

        self.store_data(trial)

        acc, mcc = self.evaluator.evaluate(validation=True)

        return mcc



from . import Generator
from .tabu.tabula import Tab
from .tabump.tabula import TabMP
from gentab.utils import console, PROG_COLUMNS, DEVICE

import os
import pandas as pd
import torch
import typing as tp
from huggingface_hub import hf_hub_download

REPO_ID = "omaralvarez/tabula"
FILENAME = "model.pt"


class Tabula(Generator):
    """Tabula Class

    The Tabula class handles the whole generation flow. It is used to fine-tune a large language model for tabular data,
    and to sample synthetic tabular data.

    Attributes:
        llm (str): HuggingFace checkpoint of a pretrained large language model, used a basis of our model
        tokenizer (AutoTokenizer): Tokenizer, automatically downloaded from llm-checkpoint
        model (AutoModelForCausalLM): Large language model, automatically downloaded from llm-checkpoint
        experiment_dir (str): Directory, where the training checkpoints will be saved
        epochs (int): Number of epochs to fine-tune the model
        batch_size (int): Batch size used for fine-tuning
        train_hyperparameters (dict): Additional hyperparameters added to the TrainingArguments used by the
         HuggingFaceLibrary, see here the full list of all possible values
         https://huggingface.co/docs/transformers/main/en/main_classes/trainer#transformers.TrainingArguments
        columns (list): List of all features/columns of the tabular dataset
        num_cols (list): List of all numerical features/columns of the tabular dataset
        categorical_columns (dict | list): only for int that are categories? did not work with conditional_col
        maybe only for other categorical labels
    """

    def __init__(
        self,
        dataset,
        llm: str = "distilgpt2",
        experiment_dir: str = "trainer_tabula",
        epochs: int = 15,
        batch_size: int = 32,
        max_tries_per_batch: int = 4096,
        resume_from_checkpoint: tp.Union[bool, str] = False,
        encode_categories: bool = False,  # Reduce token length using int categories
        # Generation options
        start_col: tp.Optional[str] = "",
        start_col_dist: tp.Optional[tp.Union[dict, list]] = None,
        temperature: float = 0.6,
        k: int = 100,
        max_length: int = 1024,
        device: str = DEVICE,
        trained_model: str = None,
        n_samples: int = 8192,
        middle_padding: bool = False,
        random_initialization: bool = False,
    ) -> None:
        super().__init__(dataset, batch_size, max_tries_per_batch)
        self.data = self.dataset.get_single_df()
        self.categorical_columns = (
            self.config["categorical_columns"] if encode_categories else []
        )
        self.llm = llm
        self.experiment_dir = experiment_dir
        self.resume_from_checkpoint = resume_from_checkpoint
        self.epochs = epochs
        self.start_col = start_col
        self.start_col_dist = start_col_dist
        self.temperature = temperature
        self.k = k
        self.max_length = max_length
        self.device = device
        self.trained_model = trained_model
        self.n_samples = n_samples

        if middle_padding:
            self.model = TabMP(
                llm=self.llm,
                experiment_dir=self.experiment_dir,
                batch_size=self.batch_size,
                epochs=self.epochs,
                categorical_columns=self.categorical_columns,
            )
        else:
            self.model = Tab(
                llm=self.llm,
                experiment_dir=self.experiment_dir,
                batch_size=self.batch_size,
                epochs=self.epochs,
                categorical_columns=self.categorical_columns,
            )

        if self.trained_model is None:
            if not middle_padding and not random_initialization:
                path = hf_hub_download(repo_id=REPO_ID, filename=FILENAME)
                self.model.model.load_state_dict(torch.load(path), strict=False)
        else:
            self.model.model.load_state_dict(
                torch.load(self.trained_model), strict=False
            )

    def preprocess(self) -> None:
        return super().preprocess()

    def train(self) -> None:
        # Setup progress
        self.p.columns = PROG_COLUMNS
        self.p.update(self.gen_task, total=self.epochs)

        if self.trained_model is None:
            self.model.fit(
                data=self.data,
                conditional_col=self.dataset.config["y_label"],
                resume_from_checkpoint=self.resume_from_checkpoint,
                progress=self.p,
                task=self.gen_task,
            )
            torch.save(
                self.model.model.state_dict(),
                os.path.join(self.experiment_dir, "model_" + str(self.dataset) + ".pt"),
            )
        else:
            self.model.init(
                data=self.data,
                conditional_col=self.dataset.config["y_label"],
            )

    def sample(self) -> pd.DataFrame:
        sample_task = self.p.add_task(
            total=self.n_samples,
            description="Generating {} batch...".format(self),
        )
        gen = self.model.sample(
            n_samples=self.n_samples,
            start_col=self.start_col,
            start_col_dist=self.start_col_dist,
            temperature=self.temperature,
            k=self.k,
            max_length=self.max_length,
            device=self.device,
            max_tries=self.max_tries_per_batch,
            progress=self.p,
            task=sample_task,
        )
        self.p.update(sample_task, visible=False)

        return gen



from gentab.evaluators import KNN, LightGBM, CatBoost, XGBoost, MLP
from gentab.generators import (
    ROS,
    SMOTE,
    ADASYN,
    TVAE,
    CTGAN,
    GaussianCopula,
    CopulaGAN,
    CTABGAN,
    CTABGANPlus,
    AutoDiffusion,
    ForestDiffusion,
    Tabula,
    GReaT,
)
from gentab.data import Config, Dataset
from gentab.utils import console

import pandas as pd
from functools import reduce


def preproc_playnet(dataset):
    dataset.reduce_size(
        {
            "left_attack": 0.97,
            "right_attack": 0.97,
            "right_transition": 0.9,
            "left_transition": 0.9,
            "time_out": 0.8,
            "left_penal": 0.5,
            "right_penal": 0.5,
        }
    )
    dataset.merge_classes(
        {
            "attack": ["left_attack", "right_attack"],
            "transition": ["left_transition", "right_transition"],
            "penalty": ["left_penal", "right_penal"],
        }
    )
    dataset.reduce_mem()

    return dataset


def preproc_adult(dataset):
    dataset.merge_classes({"<=50K": ["<=50K."], ">50K": [">50K."]})

    return dataset


def preproc_car_eval_4(dataset):
    return dataset


def evaluate_metrics(g, generator, e, df):
    evaluator = e[0](generator)
    evaluator.evaluate()

    df.loc[g[2]] = (
        evaluator.mcc,
        evaluator.accuracy,
        evaluator.weighted[0],
        evaluator.macro[0],
        evaluator.weighted[1],
        evaluator.macro[1],
        evaluator.weighted[2],
        evaluator.macro[2],
    )

    return evaluator


def get_metrics(dataset, gens, evals, metrics):
    results = []

    for e in evals:
        data = pd.DataFrame(columns=metrics)
        for g in gens:
            generator = g[0](dataset)
            try:
                try:
                    generator.load_from_disk(str(e[2]))
                    evaluator = evaluate_metrics(g, generator, e, data)
                except FileNotFoundError:
                    generator.load_from_disk()
                    evaluator = evaluate_metrics(g, generator, e, data)
            except FileNotFoundError:
                data.loc[g[2]] = (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0)

        evaluator.evaluate_baseline()
        data.loc["Baseline"] = (
            evaluator.mcc,
            evaluator.accuracy,
            evaluator.weighted[0],
            evaluator.macro[0],
            evaluator.weighted[1],
            evaluator.macro[1],
            evaluator.weighted[2],
            evaluator.macro[2],
        )

        results.append(data)

    console.print(results)

    averages = reduce(lambda x, y: x + y, results)
    averages /= len(evals)
    averages.loc[:, averages.columns != "MCC"] *= 100.0

    maxs = averages.max()
    maxs[maxs.index != "MCC"] = maxs[maxs.index != "MCC"].round(1)
    maxs.loc["MCC"] = maxs.loc["MCC"].round(2)

    averages.loc[:, averages.columns != "MCC"] = averages.loc[
        :, averages.columns != "MCC"
    ].round(1)
    averages["MCC"] = averages["MCC"].round(2)

    return averages, maxs


def get_latex_str(metric, gen, avs, mxs, round):
    avg = avs[metric][gen]
    if avg == mxs[metric]:
        return "\\textbf{{{:.{prec}f}}}".format(avg, prec=round)
    else:
        return "{:.{prec}f}".format(avg, prec=round)


def get_latex(averages, maxs, gens):
    lines = []
    line = (
        "None"
        + " & "
        + get_latex_str("MCC", "Baseline", averages, maxs, 2)
        + " & "
        + get_latex_str("Acc", "Baseline", averages, maxs, 1)
        + "\\%  & "
        + get_latex_str("WPr", "Baseline", averages, maxs, 1)
        + "\\% & "
        + get_latex_str("MPr", "Baseline", averages, maxs, 1)
        + "\\% & "
        + get_latex_str("WRe", "Baseline", averages, maxs, 1)
        + "\\% & "
        + get_latex_str("MRe", "Baseline", averages, maxs, 1)
        + "\\%  & "
        + get_latex_str("WF", "Baseline", averages, maxs, 1)
        + "\\% & "
        + get_latex_str("MF", "Baseline", averages, maxs, 1)
        + "\\% \\\\"
    )
    lines.append(line)
    line = " & "

    for g in gens:
        if averages.loc[g[2]].sum() == 0.0:
            line = g[1] + " & - & - & - & - & - & - & - & - \\\\"
        else:
            line = (
                g[1]
                + " & "
                + get_latex_str("MCC", g[2], averages, maxs, 2)
                + " & "
                + get_latex_str("Acc", g[2], averages, maxs, 1)
                + "\\%  & "
                + get_latex_str("WPr", g[2], averages, maxs, 1)
                + "\\% & "
                + get_latex_str("MPr", g[2], averages, maxs, 1)
                + "\\% & "
                + get_latex_str("WRe", g[2], averages, maxs, 1)
                + "\\% & "
                + get_latex_str("MRe", g[2], averages, maxs, 1)
                + "\\%  & "
                + get_latex_str("WF", g[2], averages, maxs, 1)
                + "\\% & "
                + get_latex_str("MF", g[2], averages, maxs, 1)
                + "\\% \\\\"
            )
        lines.append(line)

    return lines


configs = [
    ("configs/playnet.json", preproc_playnet, "PlayNet"),
    ("configs/adult.json", preproc_adult, "Adult"),
    ("configs/car_eval_4.json", preproc_car_eval_4, "Car Evaluation"),
]

gens = [
    (SMOTE, "SMOTE \cite{chawla2002smote}", "SMOTE"),
    (ADASYN, "ADASYN \cite{he2008adasyn}", "ADASYN"),
    (TVAE, "TVAE \cite{xu2019modeling}", "TVAE"),
    (CTGAN, "CTGAN \cite{xu2019modeling}", "CTGAN"),
    (GaussianCopula, "GaussianCopula \cite{patki2016synthetic}", "GaussianCopula"),
    (CopulaGAN, "CopulaGAN \cite{xu2019modeling}", "CopulaGAN"),
    (CTABGAN, "CTAB-GAN \cite{zhao2021ctab}", "CTAB-GAN"),
    (CTABGANPlus, "CTAB-GAN+ \cite{zhao2022ctab}", "CTAB-GAN+"),
    (AutoDiffusion, "AutoDiffusion \cite{suh2023autodiff}", "AutoDiffusion"),
    (
        ForestDiffusion,
        "ForestDiffusion \cite{jolicoeur2023generating}",
        "ForestDiffusion",
    ),
    (GReaT, "GReaT \cite{borisov2022language}", "GReaT"),
    (Tabula, "Tabula \cite{zhao2023tabula}", "Tabula"),
]

evals = [
    (
        LightGBM,
        "\multirow{" + str(len(gens)) + "}{*}{LightGBM \cite{ke2017lightgbm}}",
        "LightGBM",
    ),
    (
        CatBoost,
        "\multirow{"
        + str(len(gens))
        + "}{*}{CatBoost \cite{prokhorenkova2018catboost}}",
        "CatBoost",
    ),
    (
        XGBoost,
        "\multirow{" + str(len(gens)) + "}{*}{XGBoost \cite{chen2016xgboost}}",
        "XGBoost",
    ),
    (
        MLP,
        "\multirow{" + str(len(gens)) + "}{*}{MLP \cite{gorishniy2021revisiting}}",
        "MLP",
    ),
]

metrics = [
    "MCC",
    "Acc",
    "WPr",
    "MPr",
    "WRe",
    "MRe",
    "WF",
    "MF",
]

for c in configs:
    dataset = c[1](Dataset(Config(c[0])))

    averages, maxs = get_metrics(dataset, gens, evals, metrics)
    latex = get_latex(averages, maxs, gens)

    for line in latex:
        console.print(line)



from . import Generator
from gentab.utils import console

from ForestDiffusion import ForestDiffusionModel
import pandas as pd
import os
import numpy as np

os.environ["JOBLIB_TEMP_FOLDER"] = "/tmp"


class ForestDiffusion(Generator):
    def __init__(
        self,
        dataset,
        n_t=50,  # number of noise level
        model="xgboost",  # xgboost, random_forest, lgbm, catboost
        diffusion_type="flow",  # vp, flow (flow is better, but only vp can be used for imputation)
        max_depth=7,
        n_estimators=100,
        eta=0.3,  # xgboost hyperparameters
        tree_method="hist",
        reg_lambda=0.0,
        reg_alpha=0.0,
        subsample=1.0,  # xgboost hyperparameters
        num_leaves=31,  # lgbm hyperparameters
        duplicate_K=4,  # number of different noise sample per real data sample
        true_min_max_values=None,  # Vector of form [[min_x, min_y], [max_x, max_y]]; If  provided, we use these values as the min/max for each variables when using clipping
        gpu_hist=False,  # using GPU or not with xgboost
        eps=1e-3,
        beta_min=0.1,
        beta_max=8,
        n_jobs=-1,  # cpus used (feel free to limit it to something small, this will leave more cpus per model; for lgbm you have to use n_jobs=1, otherwise it will never finish)
        batch_size=8192,
        max_tries_per_batch=4096,
        **xgboost_kwargs,
    ) -> None:
        super().__init__(dataset, batch_size, max_tries_per_batch)
        self.n_t = n_t
        self.model = model
        self.diffusion_type = diffusion_type
        self.max_depth = max_depth
        self.tree_method = tree_method
        self.num_leaves = num_leaves
        self.n_estimators = n_estimators
        self.eta = eta
        self.reg_lambda = reg_lambda
        self.reg_alpha = reg_alpha
        self.subsample = subsample
        self.duplicate_K = duplicate_K
        self.true_min_max_values = true_min_max_values
        self.gpu_hist = gpu_hist
        self.eps = eps
        self.beta_min = beta_min
        self.beta_max = beta_max
        self.n_jobs = n_jobs
        # if >0 use the data iterator with the specified number of batches
        self.n_batch = batch_size

    def preprocess(self) -> None:
        X = self.dataset.encode_categories(self.dataset.X)

        # due to forestdiffusion not using pd.DF we need to get col idx...
        # vector which indicates which column is binary
        self.bin_indexes = [X.columns.get_loc(k) for k in self.config["binary_columns"]]
        # vector which indicates which column is categorical (>=3 categories)
        self.cat_indexes = [
            X.columns.get_loc(k) for k in self.config["categorical_columns"]
        ]
        # vector which indicates which column is an integer (ordinal variables such as number of cats in a box)
        self.int_indexes = [
            X.columns.get_loc(k) for k in self.config["integer_columns"]
        ]

        self.X = X.to_numpy()

        if self.dataset.config["task_type"] in ["binary", "multiclass"]:
            self.codes = pd.Categorical(self.dataset.y[self.dataset.config["y_label"]])
            self.y = self.codes.codes

        self.Xy = np.concatenate((self.X, np.expand_dims(self.y, axis=1)), axis=1)

    def train(self) -> None:
        if self.dataset.config["task_type"] in ["binary", "multiclass"]:
            self.forest_model = ForestDiffusionModel(
                self.X,
                label_y=self.y,
                n_t=self.n_t,
                model=self.model,
                diffusion_type=self.diffusion_type,
                max_depth=self.max_depth,
                tree_method=self.tree_method,
                num_leaves=self.num_leaves,
                n_estimators=self.n_estimators,
                eta=self.eta,
                reg_lambda=self.reg_lambda,
                reg_alpha=self.reg_alpha,
                subsample=self.subsample,
                duplicate_K=self.duplicate_K,
                bin_indexes=self.bin_indexes,
                cat_indexes=self.cat_indexes,
                int_indexes=self.int_indexes,
                true_min_max_values=self.true_min_max_values,
                gpu_hist=self.gpu_hist,
                eps=self.eps,
                beta_min=self.beta_min,
                beta_max=self.beta_max,
                seed=self.seed,
                n_jobs=self.n_jobs,
                n_batch=self.n_batch,
            )
        else:
            self.forest_model = ForestDiffusionModel(
                self.Xy,
                n_t=self.n_t,
                model=self.model,
                diffusion_type=self.diffusion_type,
                max_depth=self.max_depth,
                tree_method=self.tree_method,
                num_leaves=self.num_leaves,
                n_estimators=self.n_estimators,
                eta=self.eta,
                reg_lambda=self.reg_lambda,
                reg_alpha=self.reg_alpha,
                subsample=self.subsample,
                duplicate_K=self.duplicate_K,
                bin_indexes=self.bin_indexes,
                cat_indexes=self.cat_indexes,
                int_indexes=self.int_indexes,
                true_min_max_values=self.true_min_max_values,
                gpu_hist=self.gpu_hist,
                eps=self.eps,
                beta_min=self.beta_min,
                beta_max=self.beta_max,
                seed=self.seed,
                n_jobs=self.n_jobs,
                n_batch=self.n_batch,
            )

    def sample(self) -> pd.DataFrame:
        Xy_gen = self.forest_model.generate(batch_size=self.batch_size)

        df = pd.DataFrame(data=Xy_gen[:, :-1], columns=self.dataset.X.columns)

        # This is due to the library not converting back to integer binary
        # and categorical variables and not converting back to strings on its own
        int_cls = self.config["binary_columns"] + self.config["categorical_columns"]
        df[int_cls] = df[int_cls].apply(lambda col: col.round().astype("int64"))
        df = self.dataset.decode_categories(df)

        # TODO Use decode labels in dataset
        if self.dataset.config["task_type"] in ["binary", "multiclass"]:
            df[self.dataset.config["y_label"]] = pd.DataFrame({"tmp": Xy_gen[:, -1]})[
                "tmp"
            ].apply(lambda x: self.codes.categories[int(x)])
        else:
            df[self.dataset.config["y_label"]] = Xy_gen[:, -1]

        return df



from . import Tuner
from gentab.generators import Tabula
from gentab.evaluators import Evaluator
from gentab.utils import console, SPINNER, REFRESH

import optuna


class TabulaTuner(Tuner):
    def __init__(
        self,
        evaluator: Evaluator,
        trials: int,
        *args,
        min_epochs: int = 200,
        max_epochs: int = 800,
        min_batch: int = 4,
        max_batch: int = 8,
        max_tries_per_batch: int = 8192,
        timeout: int = None,
        **kwargs,
    ) -> None:
        super().__init__(
            evaluator,
            trials,
            min_epochs=min_epochs,
            max_epochs=max_epochs,
            min_batch=min_batch,
            max_batch=max_batch,
            max_tries_per_batch=max_tries_per_batch,
            timeout=timeout,
        )

    def objective(self, trial: optuna.trial.Trial) -> float:
        epochs = trial.suggest_int("epochs", self.min_epochs, self.max_epochs)
        batch_size = trial.suggest_categorical("batch_size", self.batch_sizes)
        temperature = trial.suggest_float("temperature", 0.3, 1.0)
        k = trial.suggest_int("k", 50, 200, step=2)
        encode_categories = trial.suggest_categorical(
            "encode_categories", [False, True]
        )
        middle_padding = trial.suggest_categorical("middle_padding", [False])
        random_initialization = trial.suggest_categorical(
            "random_initialization", [False, True]
        )

        self.generator = Tabula(
            self.dataset,
            epochs=epochs,
            encode_categories=encode_categories,
            middle_padding=middle_padding,
            random_initialization=random_initialization,
            batch_size=batch_size,
            temperature=temperature,
            k=k,
            max_tries_per_batch=self.max_tries_per_batch,
        )
        self.generator.generate()

        self.store_data(trial)

        acc, mcc = self.evaluator.evaluate(validation=True)

        return mcc



import numpy as np
from operator import lt, gt


class AverageMeter(object):
    def __init__(self, mv_size=128, init_val=None):
        self.mv_size = mv_size
        self.reset(init_val)

    def reset(self, init_val=None):
        self.val = [] if not init_val else [init_val]

    def update(self, val):
        self.val.append(val)

    @property
    def avg(self):
        return np.mean(self.val[-self.mv_size:])

    @property
    def overall_avg(self):
        return np.mean(self.val)


class EarlyStopping(object):
    """Monitoring an metric, flag when to stop training."""
    def __init__(self, mode='min', min_delta=0, percentage=False, patience=10, initial_bad=0, initial_best=np.nan, verbose=0):
        assert patience > 0, 'patience must be positive integer'
        assert mode in ['min', 'max'], 'mode must be either min or max'
        self.mode = mode
        self.patience = patience
        self.best = float('inf') if mode == 'min' else float('-inf')
        self.num_bad_epochs = initial_bad
        self.is_better = self._init_is_better(mode, min_delta, percentage)
        self.verbose = verbose
        self._stop = False

    def step(self, metric):
        if self.is_better(metric, self.best):
            self.num_bad_epochs = 0
            self.best = metric
        else:
            self.num_bad_epochs += 1

        if np.isnan(self.best) and (not np.isnan(metric)):
            self.num_bad_epochs = 0
            self.best = metric

        self._stop = self.num_bad_epochs >= self.patience
        if self.verbose and self._stop: print('Early Stopping Triggered, best score is: ', self.best)
        return self._stop

    def _init_is_better(self, mode, min_delta, percentage):
        comparator = lt if mode == 'min' else gt
        if not percentage:
            def _is_better(new, best):
                target = best - min_delta if mode == 'min' else best + min_delta
                return comparator(new, target)
        else:
            def _is_better(new, best):
                target = best * (1 - (min_delta / 100)) if mode == 'min' else best * (1 + (min_delta / 100))
                return comparator(new, target)
        return _is_better

    def __repr__(self):
        return '<EarlyStopping object with: mode - {} - num_bad_epochs - {} - patience - {} - best - {}>'.format(
            self.mode, self.num_bad_epochs, self.patience, self.best)



from gentab.generators import (
    SMOTE,
    ADASYN,
    TVAE,
    CTGAN,
    GaussianCopula,
    CopulaGAN,
    CTABGAN,
    CTABGANPlus,
    AutoDiffusion,
    ForestDiffusion,
    Tabula,
    GReaT,
)
from gentab.evaluators import LightGBM, XGBoost, CatBoost, MLP
from gentab.tuners import (
    SMOTETuner,
    ADASYNTuner,
    TVAETuner,
    CTGANTuner,
    GaussianCopulaTuner,
    CopulaGANTuner,
    CTABGANTuner,
    CTABGANPlusTuner,
    AutoDiffusionTuner,
    ForestDiffusionTuner,
    TabulaTuner,
    GReaTTuner,
)

from gentab.data import Config, Dataset
from gentab.utils import console


def preproc_playnet(dataset):
    dataset.reduce_size(
        {
            "left_attack": 0.97,
            "right_attack": 0.97,
            "right_transition": 0.9,
            "left_transition": 0.9,
            "time_out": 0.8,
            "left_penal": 0.5,
            "right_penal": 0.5,
        }
    )
    dataset.merge_classes(
        {
            "attack": ["left_attack", "right_attack"],
            "transition": ["left_transition", "right_transition"],
            "penalty": ["left_penal", "right_penal"],
        }
    )
    dataset.reduce_mem()

    return dataset


def preproc_adult(dataset):
    dataset.merge_classes({"<=50K": ["<=50K."], ">50K": [">50K."]})

    return dataset


def preproc_car_eval_4(dataset):
    return dataset


def tune_and_save(dataset, gen, eval, tun):
    console.print(dataset.class_counts(), dataset.row_count())
    generator = gen(dataset)
    evaluator = eval(generator)
    tuner = tun(evaluator, trials)
    tuner.tune()
    tuner.save_to_disk()
    console.print(dataset.generated_class_counts(), dataset.generated_row_count())


trials = 10

configs = [
    ("configs/playnet.json", preproc_playnet, "PlayNet"),
    ("configs/adult.json", preproc_adult, "Adult"),
    ("configs/car_eval_4.json", preproc_car_eval_4, "Car Evaluation"),
]

gens = [
    (SMOTE, "SMOTE", SMOTETuner),
    (ADASYN, "ADASYN", ADASYNTuner),
    (TVAE, "TVAE", TVAETuner),
    (CTGAN, "CTGAN", CTGANTuner),
    (GaussianCopula, "GaussianCopula", GaussianCopulaTuner),
    (CopulaGAN, "CopulaGAN", CopulaGANTuner),
    (CTABGAN, "CTAB-GAN", CTABGANTuner),
    (CTABGANPlus, "CTAB-GAN+", CTABGANPlusTuner),
    (AutoDiffusion, "AutoDiffusion", AutoDiffusionTuner),
    (ForestDiffusion, "ForestDiffusion", ForestDiffusionTuner),
    (GReaT, "GReaT", GReaTTuner),
    (Tabula, "Tabula", TabulaTuner),
]

evals = [
    (LightGBM, "LightGBM"),
    (XGBoost, "XGBoost"),
    (CatBoost, "CatBoost"),
    (MLP, "MLP"),
]

for c in configs:
    console.print(c[2])
    dataset = c[1](Dataset(Config(c[0])))

    for g in gens:
        for e in evals:
            tune_and_save(dataset, g[0], e[0], g[2])



from gentab.generators import (
    ROS,
    SMOTE,
    ADASYN,
    TVAE,
    CTGAN,
    GaussianCopula,
    CopulaGAN,
    CTABGAN,
    CTABGANPlus,
    AutoDiffusion,
    ForestDiffusion,
    Tabula,
    GReaT,
)
from gentab.data import Config, Dataset
from gentab.utils import console

config = Config("configs/car_evaluation_cr.json")

dataset = Dataset(config)

n_samples = dataset.class_counts().to_dict()

console.print(dataset.class_counts(), dataset.row_count())
generator = SMOTE(dataset)
generator.generate(n_samples=n_samples, append=False)
dataset.save_to_disk(generator)
console.print(dataset.generated_class_counts(), dataset.generated_row_count())

console.print(dataset.class_counts(), dataset.row_count())
generator = ADASYN(dataset, sampling_strategy="minority")
# generator.generate({"right_transition": 83, "time_out": 153})
generator.generate(n_samples=n_samples, append=False)
dataset.save_to_disk(generator)
console.print(dataset.generated_class_counts(), dataset.generated_row_count())

console.print(dataset.class_counts(), dataset.row_count())
generator = TVAE(dataset)
generator.generate(n_samples=n_samples, append=False)
dataset.save_to_disk(generator)
console.print(dataset.generated_class_counts(), dataset.generated_row_count())

console.print(dataset.class_counts(), dataset.row_count())
generator = CTGAN(dataset)
generator.generate(n_samples=n_samples, append=False)
dataset.save_to_disk(generator)
console.print(dataset.generated_class_counts(), dataset.generated_row_count())

console.print(dataset.class_counts(), dataset.row_count())
generator = GaussianCopula(dataset)
generator.generate(n_samples=n_samples, append=False)
dataset.save_to_disk(generator)
console.print(dataset.generated_class_counts(), dataset.generated_row_count())

console.print(dataset.class_counts(), dataset.row_count())
generator = CopulaGAN(dataset)
generator.generate(n_samples=n_samples, append=False)
dataset.save_to_disk(generator)
console.print(dataset.generated_class_counts(), dataset.generated_row_count())

console.print(dataset.class_counts(), dataset.row_count())
generator = CTABGAN(
    dataset,
    test_ratio=0.1,
    epochs=1000,
)
generator.generate(n_samples=n_samples, append=False)
console.print(dataset.generated_class_counts(), dataset.generated_row_count())
dataset.save_to_disk(generator)

console.print(dataset.class_counts(), dataset.row_count())
generator = CTABGANPlus(
    dataset,
    test_ratio=0.2,
    epochs=2000,
)
generator.generate(n_samples=n_samples, append=False)
console.print(dataset.generated_class_counts(), dataset.generated_row_count())
dataset.save_to_disk(generator)

console.print(dataset.class_counts(), dataset.row_count())
generator = AutoDiffusion(dataset)
generator.generate(n_samples=n_samples, append=False)
console.print(dataset.generated_class_counts(), dataset.generated_row_count())
dataset.save_to_disk(generator)

console.print(dataset.class_counts(), dataset.row_count())
generator = ForestDiffusion(dataset, n_jobs=1, duplicate_K=4, n_estimators=100)
generator.generate(n_samples=n_samples, append=False)
console.print(dataset.generated_class_counts(), dataset.generated_row_count())
dataset.save_to_disk(generator)

console.print(dataset.class_counts(), dataset.row_count())
generator = GReaT(
    dataset,
    epochs=600,
    max_length=1024,
    temperature=0.6,
    batch_size=32,
    max_tries_per_batch=4096,
    n_samples=8192,
)
generator.generate(n_samples=n_samples, append=False)
console.print(dataset.generated_class_counts(), dataset.generated_row_count())
dataset.save_to_disk(generator)

console.print(dataset.class_counts(), dataset.row_count())
generator = Tabula(
    dataset,
    # categorical_columns=[dataset.config["y_label"]],
    epochs=600,
    max_length=1024,
    temperature=0.6,
    batch_size=32,
    max_tries_per_batch=4096,
    n_samples=8192,
)
generator.generate(n_samples=n_samples, append=False)
console.print(dataset.generated_class_counts(), dataset.generated_row_count())
dataset.save_to_disk(generator)



from . import Tuner
from gentab.generators import ADASYN
from gentab.evaluators import Evaluator
from gentab.utils import console, SPINNER, REFRESH

import optuna


class ADASYNTuner(Tuner):
    def __init__(
        self,
        evaluator: Evaluator,
        trials: int,
        *args,
        timeout: int = None,
        **kwargs,
    ) -> None:
        super().__init__(evaluator, trials, timeout=timeout)

    def objective(self, trial: optuna.trial.Trial) -> float:
        n_neighbors = trial.suggest_int(
            "n_neighbors", 2, self.dataset.min_class_count(), log=True
        )

        self.generator = ADASYN(self.dataset, n_neighbors=n_neighbors)
        self.generator.generate()

        self.store_data(trial)

        acc, mcc = self.evaluator.evaluate(validation=True)

        return mcc



import torch
import numpy as np
from torch import nn
from torch.nn import functional as F
from einops import rearrange


__all__ = ['SwapNoiseCorrupter', 'AutoEncoder']


class SwapNoiseCorrupter(nn.Module):
    """
        Apply swap noise on the input data.

        Each data point has specified chance be replaced by a random value from the same column.
    """
    def __init__(self, probas):
        super().__init__()
        self.probas = torch.from_numpy(np.array(probas))

    def forward(self, x):
        should_swap = torch.bernoulli(self.probas.to(x.device) * torch.ones((x.shape)).to(x.device))
        corrupted_x = torch.where(should_swap == 1, x[torch.randperm(x.shape[0])], x)
        mask = (corrupted_x != x).float()
        return corrupted_x, mask


class EntityEmbedder(nn.Module):
    ''' Embed categorical variables. '''
    def __init__(self, cards, embeded_dims):
        super().__init__()
        assert len(cards) == len(embeded_dims), 'require embedding dim for each categorical variables'
        self.embedders = nn.ModuleList([
            nn.Embedding(c, s)
            for c, s in zip(cards, embeded_dims)
        ])

    def forward(self, x):
        return torch.cat([
            embedder(x[:, i].long())
            for i, embedder in enumerate(self.embedders)
        ], dim=1)


class OneHotEncoder(nn.Module):
    ''' One-Hot encode categorical variables. '''
    def __init__(self, cards):
        super().__init__()
        self.cards = cards

    def forward(self, x):
        return torch.cat([
            F.one_hot(x[:, i].long(), num_classes=num_classes)
            for i, num_classes in enumerate(self.cards)
        ], dim=1)


def _make_mlp_layers(num_units):
    layers = nn.ModuleList([
        nn.Linear(in_features, out_features)
        for in_features, out_features in zip(num_units, num_units[1:])
    ])
    return layers


class DeapStack(nn.Module):
    ''' Simple MLP body. '''
    def __init__(self, in_features, hidden_size, num_layers=3):
        super().__init__()
        self.layers = _make_mlp_layers([in_features] + [hidden_size] * num_layers)
        self._output_shape = hidden_size

    @property
    def output_shape(self): return self._output_shape

    def forward_pass(self, x):
        outputs = []
        for layer in self.layers:
            x = F.relu(layer(x))
            outputs.append(x)
        return outputs

    def forward(self, x):
        return self.forward_pass(x)[~0]

    def featurize(self, x):
        return torch.cat(self.forward_pass(x), dim=1)


class DeepBottleneck(nn.Module):
    ''' Simple MLP body with bottleneck. '''
    def __init__(self, in_features, hidden_size, bottleneck_size, num_layers=3):
        super().__init__()
        encoder_layers = num_layers >> 1
        decoder_layers = num_layers - encoder_layers - 1
        self.encoder = _make_mlp_layers([in_features] + [hidden_size] * encoder_layers)
        self.bottleneck = nn.Linear(hidden_size, bottleneck_size)
        self.decoder = _make_mlp_layers([bottleneck_size] + [hidden_size] * decoder_layers)
        self._output_shape = hidden_size

    @property
    def output_shape(self): return self._output_shape

    def forward_pass(self, x):
        for layer in self.encoder:
            x = F.relu(layer(x))
        x = b = self.bottleneck(x)
        for layer in self.decoder:
            x = F.relu(layer(x))
        return [b, x]

    def forward(self, x):
        return self.forward_pass(x)[1]

    def featurize(self, x):
        return self.forward_pass(x)[0]


class TransformerEncoder(nn.Module):
    ''' Transformer Encoder. '''
    def __init__(self, embed_dim, num_heads, dropout, feedforward_dim):
        super().__init__()
        assert embed_dim % num_heads == 0, 'embed_dim must be a multiple of num_heads'
        self.attn = nn.MultiheadAttention(embed_dim, num_heads, dropout=dropout)
        self.linear_1 = nn.Linear(embed_dim, feedforward_dim)
        self.linear_2 = nn.Linear(feedforward_dim, embed_dim)
        self.layernorm_1 = nn.LayerNorm(embed_dim)
        self.layernorm_2 = nn.LayerNorm(embed_dim)

    def forward(self, x_in):
        ''' input is of shape num_subspaces x batch_size x embed_dim '''
        attn_out, _ = self.attn(x_in, x_in, x_in)
        x = self.layernorm_1(x_in + attn_out)
        ff_out = self.linear_2(F.relu(self.linear_1(x)))
        x = self.layernorm_2(x + ff_out)
        return x


class Transformer(nn.Module):
    ''' DAE Body with transformer encoders. '''
    def __init__(self, in_features, hidden_size=1024, num_subspaces=8, embed_dim=128, num_heads=8, dropout=0, feedforward_dim=512, num_layers=3):
        super().__init__()
        assert hidden_size == embed_dim * num_subspaces, 'num_subspaces must be a multiple of embed_dim'
        self.num_subspaces = num_subspaces
        self.embed_dim = embed_dim

        self.excite = nn.Linear(in_features, hidden_size)
        self.encoders = nn.ModuleList([
            TransformerEncoder(embed_dim, num_heads, dropout, feedforward_dim)
            for _ in range(num_layers)
        ])
        self._output_shape = hidden_size

    @property
    def output_shape(self): return self._output_shape

    def divide(self, x):
        return rearrange(x, 'b (s e) -> s b e', s=self.num_subspaces, e=self.embed_dim)

    def combine(self, x):
        return rearrange(x, 's b e -> b (s e)')

    def forward_pass(self, x):
        x = F.relu(self.excite(x))
        x = self.divide(x)
        outputs = []
        for encoder in self.encoders:
            x = encoder(x)
            outputs.append(x)
        return outputs

    def forward(self, x):
        return self.combine(self.forward_pass(x)[~0])

    def featurize(self, x):
        return torch.cat([self.combine(x) for x in self.forward_pass(x)], dim=1)


class MultiTaskHead(nn.Module):
    """
        Simple Linear transformation to take last hidden representation to reconstruct inputs.

        Output is dictionary of variable type to tensor mapping.
    """
    def __init__(self, in_features, n_bins=0, n_cats=0, n_nums=0, cards=[]):
        super().__init__()
        assert n_cats == len(cards), 'require cardinalities for each categorical variable'
        assert n_bins + n_cats + n_nums, 'need some targets'
        self.n_bins = n_bins
        self.n_cats = n_cats
        self.n_nums = n_nums

        self.bins_linear = nn.Linear(in_features, n_bins) if n_bins else None
        self.cats_linears = nn.ModuleList([nn.Linear(in_features, card) for card in cards])
        self.nums_linear = nn.Linear(in_features, n_nums) if n_nums else None

    def forward(self, features):
        outputs = dict()

        if self.bins_linear:
            outputs['bins'] = self.bins_linear(features)

        if self.cats_linears:
            outputs['cats'] = [linear(features) for linear in self.cats_linears]

        if self.nums_linear:
            outputs['nums'] = self.nums_linear(features)

        return outputs


_ae_body_options = {
    'deepstack': DeapStack,
    'deepbottleneck': DeepBottleneck,
    'transformer': Transformer
}


class AutoEncoder(nn.Module):
    ''' AutoEncoder. '''
    def __init__(self, datatype_info, cards=[], embeded_dims=[], cats_handling='onehot',
                       body_network=DeapStack, body_network_cfg=dict()):
        super().__init__()
        self.n_bins = n_bins = datatype_info.get('n_bins', 0)
        self.n_cats = n_cats = datatype_info.get('n_cats', 0)
        self.n_nums = n_nums = datatype_info.get('n_nums', 0)
        self.cards = cards
        self.embeded_dims = embeded_dims if isinstance(embeded_dims, (list, tuple, np.ndarray)) else [embeded_dims] * n_cats

        if isinstance(body_network, str):
            body_network = _ae_body_options[body_network]

        self.body_network_in_features = n_bins + n_nums

        self.cats_handler = None
        if n_cats:
            if cats_handling == 'onehot':
                self.cats_handler = OneHotEncoder(cards)
                self.body_network_in_features += sum(cards)
            if cats_handling == 'embed':
                self.cats_handler = EntityEmbedder(cards, self.embeded_dims)
                self.body_network_in_features += sum(self.embeded_dims)

        self.body = body_network(in_features=self.body_network_in_features, **body_network_cfg)
        self.reconstruction_head = MultiTaskHead(self.body.output_shape, n_bins, n_cats, n_nums, cards)
        self.mask_predictor_head = nn.Linear(self.body.output_shape, sum([n_bins, n_cats, n_nums]))

    def process_inputs(self, inputs):
        """
            Process inputs to the network.

            Input is dictionary of variable type to tensor mapping.
            Output is 2D tensor of size: batch_size x body_network_in_features.
        """
        x_bins = inputs.get('bins', None)
        x_cats = inputs.get('cats', None)
        x_nums = inputs.get('nums', None)
        if x_cats is not None: x_cats = self.cats_handler(x_cats)
        body_network_inputs = torch.cat([t for t in [x_bins, x_cats, x_nums] if t is not None], dim=1)
        return body_network_inputs

    def forward(self, inputs):
        """ Forward pass of AutoEncoder network.

            Inputs is dictionary of variable type to tensor mapping.
            Outputs are tuple of reconstruction of inputs and prediction of corruption mask.
        """
        body_network_inputs = self.process_inputs(inputs)
        last_hidden = self.body(body_network_inputs)
        reconstruction = self.reconstruction_head(last_hidden)
        predicted_mask = self.mask_predictor_head(last_hidden)
        return reconstruction, predicted_mask

    def featurize(self, inputs):
        body_network_inputs = self.process_inputs(inputs)
        features = self.body.featurize(body_network_inputs)
        return features


    def loss(self, inputs, masks, reconstruction, predicted_mask, loss_weights):
        """ Calculating the loss for DAE network.

            BCE for masks and reconstruction of binary inputs.
            CE for categoricals.
            MSE for numericals.

            reconstruction loss is weighted average of mean reduction of loss per datatype.
            mask loss is mean reduced.
            final loss is weighted sum of reconstruction loss and mask loss.
        """
        flattened_masks = torch.cat(list(masks.values()), dim=1)
        mask_loss = loss_weights['mask'] * F.binary_cross_entropy_with_logits(predicted_mask, flattened_masks)

        reconstruction_losses = dict()
        if 'bins' in inputs:
            reconstruction_losses['bins'] = loss_weights['bins'] * F.binary_cross_entropy_with_logits(reconstruction['bins'], inputs['bins'])

        if 'cats' in inputs:
            cats_losses = []
            for i in range(len(reconstruction['cats'])):
                cats_losses.append(F.cross_entropy(reconstruction['cats'][i], inputs['cats'][:, i].long()))
            reconstruction_losses['cats'] = loss_weights['cats'] * torch.stack(cats_losses).mean()

        if 'nums' in inputs:
            reconstruction_losses['nums'] = loss_weights['nums'] * F.mse_loss(reconstruction['nums'], inputs['nums'])

        reconstruction_loss = torch.stack(list(reconstruction_losses.values())).mean()
        return reconstruction_loss + mask_loss



from gentab.utils import console, ProgressBar
from gentab import SEED

import os
from pathlib import Path
from typing import Any, Literal, Optional, Union, cast, Tuple, Dict, List

import pandas as pd
import dask.dataframe as dd
import numpy as np
import torch
from sklearn.preprocessing import LabelEncoder, OneHotEncoder
from sklearn.model_selection import train_test_split
from scipy.stats import wasserstein_distance
from scipy.spatial.distance import jensenshannon
from dython.nominal import theils_u
from imblearn.datasets import fetch_datasets
from ucimlrepo import fetch_ucirepo


class Dataset:
    def __init__(self, config, cache_path="datasets") -> None:
        self.config = config
        self.cache_path = cache_path
        self.X_gen = None
        self.y_gen = None

        with ProgressBar(indeterminate=True).progress as p:
            gen_task = p.add_task(
                "Loading dataset {}...".format(self.config["name"]), total=None
            )

            if not self.config.exists("download"):
                self.load_path()
            else:
                if self.config["download"] == "ucimlrepo":
                    self.download_uci()
                else:
                    self.download_imb()

            self.get_label_encoders()
            self.compute_categories()

        console.print("✅ Dataset loaded...")

    def __str__(self) -> str:
        return self.config["name"]

    def reset_indexes(self) -> None:
        # Reset indexes
        self.X.reset_index(drop=True, inplace=True)
        self.X_val.reset_index(drop=True, inplace=True)
        self.X_test.reset_index(drop=True, inplace=True)
        self.y.reset_index(drop=True, inplace=True)
        self.y_val.reset_index(drop=True, inplace=True)
        self.y_test.reset_index(drop=True, inplace=True)

    def load_path(self) -> None:
        self.X = pd.read_csv(self.config["path_X"])
        self.y = pd.read_csv(self.config["path_y"])
        self.X_test = pd.read_csv(self.config["path_X_test"])
        self.y_test = pd.read_csv(self.config["path_y_test"])

        self.X, self.X_val, self.y, self.y_val = train_test_split(
            self.X,
            self.y,
            test_size=self.config["val_size"],
            random_state=SEED,
            stratify=self.y,
        )

    def download_imb(self) -> None:
        Path(self.cache_path).mkdir(parents=True, exist_ok=True)

        data = fetch_datasets(
            filter_data=(self.config["name"],),
            data_home=self.cache_path,
            random_state=SEED,
        )[self.config["name"]]

        features = pd.DataFrame(
            data=data.data,
            columns=["f" + str(i + 1) for i in range(data.data.shape[1])],
        )
        labels = pd.DataFrame({self.config["y_label"]: data.target})

        self.X, self.X_test, self.y, self.y_test = train_test_split(
            features,
            labels,
            test_size=1 - self.config["train_size"],
            random_state=SEED,
            stratify=labels,
        )
        self.X_val, self.X_test, self.y_val, self.y_test = train_test_split(
            self.X_test,
            self.y_test,
            test_size=self.config["test_size"]
            / (self.config["test_size"] + self.config["val_size"]),
            random_state=SEED,
            stratify=self.y_test,
        )

        self.reset_indexes()

        self.config["binary_columns"] = self.X.columns.values.tolist()
        self.config["categorical_columns"] = []
        self.config["integer_columns"] = []

    def download_uci(self) -> None:
        Path(os.path.join(self.cache_path, "uci")).mkdir(parents=True, exist_ok=True)

        path_features = os.path.join(
            self.cache_path, "uci", self.config["name"] + "_features.csv"
        )
        path_targets = os.path.join(
            self.cache_path, "uci", self.config["name"] + "_targets.csv"
        )

        if Path(path_features).is_file() and Path(path_targets).is_file():
            features = pd.read_csv(path_features)
            targets = pd.read_csv(path_targets)
        else:
            uci = fetch_ucirepo(name=self.config["name"])

            # metadata
            console.print(uci.variables)

            uci.data.features.fillna("Missing", inplace=True)

            uci.data.features.to_csv(path_features, index=False)
            uci.data.targets.to_csv(path_targets, index=False)

            features = uci.data.features
            targets = uci.data.targets

        self.X, self.X_test, self.y, self.y_test = train_test_split(
            features,
            targets,
            test_size=1 - self.config["train_size"],
            random_state=SEED,
            stratify=targets,
        )
        self.X_val, self.X_test, self.y_val, self.y_test = train_test_split(
            self.X_test,
            self.y_test,
            test_size=self.config["test_size"]
            / (self.config["test_size"] + self.config["val_size"]),
            random_state=SEED,
            stratify=self.y_test,
        )

        self.reset_indexes()

    def save_to_disk(self, generator, tuner="") -> None:
        with ProgressBar(indeterminate=True).progress as p:
            gen_task = p.add_task(
                "Saving dataset to {}...".format(self.config["save_path"]), total=None
            )

            if tuner == "":
                path = self.config["save_path"]
            else:
                path = self.config["save_path"] + "_" + str(tuner).lower()

            Path(path).mkdir(parents=True, exist_ok=True)

            self.X_gen.to_csv(
                os.path.join(path, "X_gen_" + str(generator) + ".csv"),
                index=False,
            )
            self.y_gen.to_csv(
                os.path.join(path, "y_gen_" + str(generator) + ".csv"),
                index=False,
            )

        console.print("✅ Dataset saved to {}...".format(path))

    def load_from_disk(self, generator, tuner="") -> None:
        with ProgressBar(indeterminate=True).progress as p:
            gen_task = p.add_task("Loading dataset {}...".format(generator), total=None)

            if tuner == "":
                path = self.config["save_path"]
            else:
                path = self.config["save_path"] + "_" + str(tuner).lower()

            self.X_gen = pd.read_csv(
                os.path.join(path, "X_gen_" + str(generator) + ".csv")
            )
            self.y_gen = pd.read_csv(
                os.path.join(path, "y_gen_" + str(generator) + ".csv")
            )

        console.print("✅ {} dataset loaded from {}...".format(generator, path))

    def compute_categories(self) -> None:
        self.cats = self.config["categorical_columns"] + self.config["binary_columns"]
        self.X_cats = self.X[self.cats].copy()
        self.X_cats[self.cats] = self.X_cats[self.cats].apply(
            lambda col: pd.Categorical(col)
        )

    def get_categories(self) -> list[str]:
        return self.cats

    def get_continuous(self) -> list[str]:
        return self.X.columns.difference(
            self.cats + [self.config["y_label"]]
        ).values.tolist()

    def encode_categories(self, df: pd.DataFrame) -> pd.DataFrame:
        if self.config.exists("download") and self.config["download"] == "imbalanced":
            return df
        else:
            X_enc = df.copy()
            X_enc[self.cats] = X_enc[self.cats].apply(
                lambda col: pd.Categorical(col).codes
            )

        return X_enc

    def decode_categories(self, df: pd.DataFrame) -> pd.DataFrame:
        if self.config.exists("download") and self.config["download"] == "imbalanced":
            return df
        else:
            cat_columns = self.X_cats.select_dtypes(["category"]).columns
            df[cat_columns] = df[cat_columns].apply(
                lambda col: pd.Categorical.from_codes(
                    col, self.X_cats[col.name].cat.categories
                )
            )

        return df

    def merge_classes(self, merge: dict[str, list[str]]) -> None:
        for cls, labs in merge.items():
            self.y[self.y[self.config["y_label"]].isin(labs)] = cls
            self.y_val[self.y_val[self.config["y_label"]].isin(labs)] = cls
            self.y_test[self.y_test[self.config["y_label"]].isin(labs)] = cls

        self.get_label_encoders()

    def get_label_encoders(self) -> None:
        self.label_encoder = LabelEncoder()
        self.label_encoder.fit(self.y)
        self.label_encoder_ohe = OneHotEncoder(sparse_output=False)
        self.label_encoder_ohe.fit(self.y)

    def encode_labels(self, df: pd.DataFrame) -> pd.DataFrame:
        return self.label_encoder.transform(df)

    def decode_labels(self, df: pd.DataFrame) -> pd.DataFrame:
        return self.label_encoder.inverse_transform(df)

    def num_classes(self) -> int:
        return self.y[self.config["y_label"]].nunique()

    def num_features(self) -> int:
        return len(self.X.columns)

    def get_feature_names(self) -> list[str]:
        return self.X.columns.values.tolist()

    def class_names(self) -> list[str]:
        return self.y[self.config["y_label"]].unique().tolist()

    def class_counts(self) -> int:
        return self.y[self.config["y_label"]].value_counts()

    def min_class_count(self) -> int:
        return self.y[self.config["y_label"]].value_counts().min()

    def row_count(self) -> int:
        return len(self.X.index)

    def generated_class_counts(self) -> int:
        return self.y_gen.value_counts()

    def generated_row_count(self) -> int:
        return len(self.X_gen.index)

    def get_single_df(self) -> pd.DataFrame:
        return pd.concat([self.X, self.y], axis=1)

    def set_split_result(self, data) -> None:
        self.X_gen = data.loc[:, data.columns != self.config["y_label"]]
        self.y_gen = data[[self.config["y_label"]]]

    def get_random_class_rows(self, cls: str, n: int):
        compliant_rows = self.y[self.y[self.config["y_label"]] == cls]
        idx = compliant_rows.sample(n=n, random_state=SEED).index
        return self.X.loc[idx]

    def get_random_gen_class_rows(self, cls: str, n: int):
        compliant_rows = self.y_gen[self.y_gen[self.config["y_label"]] == cls]
        idx = compliant_rows.sample(n=n, random_state=SEED).index
        return self.X_gen.loc[idx]

    def get_class_rows(self, cls: str):
        compliant_rows = self.y[self.y[self.config["y_label"]] == cls]
        idx = compliant_rows.index
        return self.X.loc[idx]

    def get_gen_class_rows(self, cls: str):
        compliant_rows = self.y_gen[self.y_gen[self.config["y_label"]] == cls]
        idx = compliant_rows.index
        return self.X_gen.loc[idx]

    def reduce_size(self, class_percentages) -> None:
        for cls, percent in class_percentages.items():
            self.reduce_uniformly_randomly(
                self.y[self.config["y_label"]] == cls, percent
            )

    def reduce_uniformly_randomly(self, condition, percentage) -> None:
        """
        Removes a random subset of rows from a DataFrame based on a condition.

        Parameters:
        - df: pandas DataFrame.
        - condition: A boolean series indicating which rows comply with the condition.
        - percentage: Percentage of rows to remove that comply with the condition.
        """
        # Identify rows that comply with the condition
        compliant_rows = self.y[condition]

        # Calculate the number of rows to remove
        n_remove = int(len(compliant_rows) * percentage)

        # Randomly select rows to remove
        rows_to_remove = compliant_rows.sample(n=n_remove, random_state=SEED).index

        # Remove the selected rows from the DataFrame
        self.X.drop(rows_to_remove, inplace=True)
        self.X.reset_index(drop=True, inplace=True)
        self.y.drop(rows_to_remove, inplace=True)
        self.y.reset_index(drop=True, inplace=True)

    def memory_usage_mb(self, df) -> float:
        return df.memory_usage().sum() / 1024**2

    def reduce_mem_df(self, df) -> None:
        for col in df.columns:
            col_type = self.X[col].dtype

            if col_type != object and col_type != "category":
                c_min = self.X[col].min()
                c_max = self.X[col].max()
                if str(col_type)[:3] == "int":
                    if c_min > np.iinfo(np.int8).min and c_max < np.iinfo(np.int8).max:
                        self.X[col] = self.X[col].astype(np.int8)
                    elif (
                        c_min > np.iinfo(np.int16).min
                        and c_max < np.iinfo(np.int16).max
                    ):
                        self.X[col] = self.X[col].astype(np.int16)
                    elif (
                        c_min > np.iinfo(np.int32).min
                        and c_max < np.iinfo(np.int32).max
                    ):
                        self.X[col] = self.X[col].astype(np.int32)
                    elif (
                        c_min > np.iinfo(np.int64).min
                        and c_max < np.iinfo(np.int64).max
                    ):
                        self.X[col] = self.X[col].astype(np.int64)
                else:
                    if (
                        c_min > np.finfo(np.float16).min
                        and c_max < np.finfo(np.float16).max
                    ):
                        self.X[col] = self.X[col].astype(np.float16)
                    elif (
                        c_min > np.finfo(np.float32).min
                        and c_max < np.finfo(np.float32).max
                    ):
                        self.X[col] = self.X[col].astype(np.float32)
                    else:
                        self.X[col] = self.X[col].astype(np.float64)
            else:
                self.X[col] = self.X[col].astype("category")

    def reduce_mem(self) -> None:
        """iterate through all the columns of a dataframe and modify the data type
        to reduce memory usage.
        """
        start_mem = (
            self.memory_usage_mb(self.X)
            + self.memory_usage_mb(self.X_val)
            + self.memory_usage_mb(self.X_test)
        )
        console.print("💾 Memory usage of dataframe is {:.2f} MB...".format(start_mem))

        with ProgressBar(indeterminate=True).progress as p:
            p.add_task("Reducing memory usage...", total=None)

            self.reduce_mem_df(self.X)
            self.reduce_mem_df(self.X_val)
            self.reduce_mem_df(self.X_test)

        end_mem = (
            self.memory_usage_mb(self.X)
            + self.memory_usage_mb(self.X_val)
            + self.memory_usage_mb(self.X_test)
        )
        console.print(
            "💾 Memory usage after optimization: {:.2f} MB...".format(end_mem)
        )
        console.print(
            "✅ Memory usage reduced by {:.1f}%...".format(
                100 * (start_mem - end_mem) / start_mem
            )
        )

    def reduce_mem_gen(self) -> None:
        """iterate through all the columns of a dataframe and modify the data type
        to reduce memory usage.
        """
        start_mem = self.memory_usage_mb(self.X_gen)
        console.print("💾 Memory usage of dataframe is {:.2f} MB...".format(start_mem))

        with ProgressBar(indeterminate=True).progress as p:
            p.add_task("Reducing memory usage...", total=None)

            self.reduce_mem_df(self.X_gen)

        end_mem = self.memory_usage_mb(self.X_gen)
        console.print(
            "💾 Memory usage after optimization: {:.2f} MB...".format(end_mem)
        )
        console.print(
            "💾 Memory usage reduced by {:.1f}%...".format(
                100 * (start_mem - end_mem) / start_mem
            )
        )

    def get_single_encoded_data(self):
        X_real = self.encode_categories(self.X)
        X_gen = self.encode_categories(self.X_gen)

        y_real = pd.Series(self.encode_labels(self.y), name=self.config["y_label"])
        y_gen = pd.Series(self.encode_labels(self.y_gen), name=self.config["y_label"])

        real_data = pd.concat([X_real, y_real], axis=1)
        gen_data = pd.concat([X_gen, y_gen], axis=1)

        return real_data, gen_data

    def theils_u_mat(self, df):
        # Compute Theil's U-statistics between each pair of columns
        cate_columns = df.shape[1]
        theils_u_mat = np.zeros((cate_columns, cate_columns))

        for i in range(cate_columns):
            for j in range(cate_columns):
                theils_u_mat[i, j] = theils_u(df.iloc[:, i], df.iloc[:, j])

        return theils_u_mat

    # See the post https://towardsdatascience.com/the-search-for-categorical-correlation-a1cf7f1888c9
    def correlation_ratio(self, categories, measurements):
        fcat, _ = pd.factorize(categories)
        cat_num = np.max(fcat) + 1
        y_avg_array = np.zeros(cat_num)
        n_array = np.zeros(cat_num)
        for i in range(0, cat_num):
            cat_measures = measurements[np.argwhere(fcat == i).flatten()]
            n_array[i] = len(cat_measures)
            y_avg_array[i] = np.average(cat_measures)
        y_total_avg = np.sum(np.multiply(y_avg_array, n_array)) / np.sum(n_array)
        numerator = np.sum(
            np.multiply(n_array, np.power(np.subtract(y_avg_array, y_total_avg), 2))
        )
        denominator = np.sum(np.power(np.subtract(measurements, y_total_avg), 2))
        if numerator == 0:
            eta = 0.0
        else:
            eta = numerator / denominator
        return eta

    def ratio_mat(self, df, continuous_columns, categorical_columns):
        rat_mat = pd.DataFrame(index=continuous_columns, columns=categorical_columns)

        if len(categorical_columns) == 0 or len(continuous_columns) == 0:
            return np.zeros(1)
        else:
            for cat_col in categorical_columns:
                for cont_col in continuous_columns:
                    rat_mat[cat_col][cont_col] = self.correlation_ratio(
                        df[cat_col], df[cont_col]
                    )
        return rat_mat.values

    def fillNa_cont(self, df):
        for col in df.columns:
            mean_values = df[col].mean()
            df[col].fillna(mean_values, inplace=True)
        return df

    def fillNa_cate(self, df):
        for col in df.columns:
            mode_values = df[col].mode()[0]
            df[col].fillna(mode_values, inplace=True)
        return df

    def compute_correlations(self, df):
        num_mat = pd.DataFrame(df[self.get_continuous()])
        cat_mat = pd.DataFrame(df[self.get_categories()])

        num_mat = self.fillNa_cont(num_mat)
        cat_mat = self.fillNa_cate(cat_mat)

        pearson_sub_matrix = np.corrcoef(num_mat, rowvar=False)
        theils_u_matrix = self.theils_u_mat(cat_mat)
        correl_ratio_mat = self.ratio_mat(
            df, self.get_continuous(), self.get_categories()
        )

        return (pearson_sub_matrix, theils_u_matrix, correl_ratio_mat)

    def get_correlations(self):
        with ProgressBar(indeterminate=True).progress as p:
            p.add_task(
                "Computing Pearson Correlation, Theil's U, and Correlation Ratio...",
                total=None,
            )
            real_data, gen_data = self.get_single_encoded_data()
            real_pear, real_theils, real_ratio = self.compute_correlations(real_data)
            gen_pear, gen_theils, gen_ratio = self.compute_correlations(gen_data)

        console.print(
            "✅ Pearson Correlation, Theil's U, and Correlation Ratio computation complete..."
        )

        return (
            np.linalg.norm(real_pear - gen_pear),
            np.linalg.norm(real_theils - gen_theils),
            np.linalg.norm(real_ratio - gen_ratio),
        )

    def get_pearson_correlation(self) -> pd.Series:
        real_data, gen_data = self.get_single_encoded_data()

        return (
            real_data.corr(method="pearson") - gen_data.corr(method="pearson")
        ).abs()

    def distance_closest_record(self):
        with ProgressBar(indeterminate=True).progress as p:
            p.add_task("Computing DCR...", total=None)

            real_data, gen_data = self.get_single_encoded_data()

            # Convert DataFrames to Dask DataFrames
            real_ddf = dd.from_pandas(
                real_data, npartitions=5
            )  # Adjust npartitions based on your available memory
            gen_ddf = dd.from_pandas(
                gen_data, npartitions=5
            )  # Adjust npartitions based on your available memory

            # Function to compute the minimum L2 distance for each row in syn_df with respect to real_df
            def compute_min_l2_distance(row, real_array):
                distance_array = np.sqrt(((row.values - real_array) ** 2).sum(axis=1))
                return np.min(distance_array)

            # Calculate the minimum L2 distance for each row in syn_df with respect to real_df
            real_array = real_ddf.compute().values
            gen_ddf["Min_L2_Distance"] = gen_ddf.map_partitions(
                lambda part: part.apply(
                    compute_min_l2_distance, axis=1, args=(real_array,)
                ),
                meta=("Min_L2_Distance", "f8"),
            )

            # Convert the Dask DataFrame to a Pandas DataFrame
            gen_df_result = gen_ddf.compute()
            min_distances = gen_df_result["Min_L2_Distance"]

        console.print("✅ DCR computation complete...")

        return min_distances

    def jensen_shannon_distance(self):
        with ProgressBar(indeterminate=True).progress as p:
            p.add_task("Computing Jensen Shannon Distance...", total=None)

            real_data, gen_data = self.get_single_encoded_data()
            real_data = real_data[self.get_categories()]
            gen_data = gen_data[self.get_categories()]

            distances = real_data.apply(lambda x: jensenshannon(x, gen_data[x.name]))

        console.print("✅ Jensen Shannon computation complete...")

        return distances

    def wasserstein_distance(self):
        with ProgressBar(indeterminate=True).progress as p:
            p.add_task("Computing Wasserstein Distance...", total=None)

            real_data, gen_data = self.get_single_encoded_data()
            real_data = real_data[self.get_continuous()]
            gen_data = gen_data[self.get_continuous()]

            distances = real_data.apply(
                lambda x: wasserstein_distance(x, gen_data[x.name])
            )

        console.print("✅ Wasserstein computation complete...")

        return distances



from gentab.evaluators import KNN, LightGBM, XGBoost, MLP
from gentab.generators import (
    ROS,
    SMOTE,
    ADASYN,
    TVAE,
    CTGAN,
    GaussianCopula,
    CopulaGAN,
    CTABGAN,
    CTABGANPlus,
    AutoDiffusion,
    ForestDiffusion,
    Tabula,
    GReaT,
)
from gentab.data import Config, Dataset
from gentab.utils import console

import pandas as pd
import numpy as np


def preproc_playnet(dataset):
    dataset.reduce_size(
        {
            "left_attack": 0.97,
            "right_attack": 0.97,
            "right_transition": 0.9,
            "left_transition": 0.9,
            "time_out": 0.8,
            "left_penal": 0.5,
            "right_penal": 0.5,
        }
    )
    dataset.merge_classes(
        {
            "attack": ["left_attack", "right_attack"],
            "transition": ["left_transition", "right_transition"],
            "penalty": ["left_penal", "right_penal"],
        }
    )
    dataset.reduce_mem()

    return dataset


def preproc_adult(dataset):
    dataset.merge_classes({"<=50K": ["<=50K."], ">50K": [">50K."]})
    dataset.reduce_mem()

    return dataset


def preproc_car(dataset):
    return dataset


configs = [
    ("configs/car_evaluation_cr.json", preproc_car, "Car Evaluation"),
    ("configs/playnet_cr.json", preproc_playnet, "PlayNet"),
    ("configs/adult_cr.json", preproc_adult, "Adult"),
]

gens = [
    (TVAE, "TVAE \cite{xu2019modeling}"),
    (CTGAN, "CTGAN \cite{xu2019modeling}"),
    (GaussianCopula, "GaussianCopula \cite{patki2016synthetic}"),
    (CopulaGAN, "CopulaGAN \cite{xu2019modeling}"),
    (CTABGAN, "CTAB-GAN \cite{zhao2021ctab}"),
    (CTABGANPlus, "CTAB-GAN+ \cite{zhao2022ctab}"),
    (AutoDiffusion, "AutoDiffusion \cite{suh2023autodiff}"),
    (ForestDiffusion, "ForestDiffusion \cite{jolicoeur2023generating}"),
    (GReaT, "GReaT \cite{borisov2022language}"),
    (Tabula, "Tabula \cite{zhao2023tabula}"),
]

DCR = pd.DataFrame()

for c in configs:
    config = Config(c[0])

    dataset = c[1](Dataset(config))
    console.print(dataset.class_counts(), dataset.row_count())

    for g in gens:
        generator = g[0](dataset)
        try:
            generator.load_from_disk()

            min_l2_dist = dataset.distance_closest_record()
            DCR.loc[c[2], g[1]] = np.mean(min_l2_dist)
        except FileNotFoundError:
            DCR.loc[c[2], g[1]] = 0.0

round = 2

console.print(DCR)

DCR_mean = DCR.mean()

console.print(DCR_mean)

DCR_ranks = DCR.rank(ascending=True, axis=1)
console.print(DCR_ranks)
DCR_mean_rank = DCR_ranks.mean().round(round)
max = DCR_mean_rank.max()
console.print(DCR_mean_rank)

lines = []
for index, row in DCR_mean_rank.items():
    if max == row:
        line = (
            index + " & " + "\\textbf{{{:.{prec}f}}}".format(row, prec=round) + " \\\\"
        )
    elif row != float("inf"):
        line = index + " & " + "{:.{prec}f}".format(row, prec=round) + " \\\\"
    else:
        line = index + " & - \\\\"

    lines.append(line)

for line in lines:
    console.print(line)



from gentab.evaluators import Evaluator
from gentab.utils import console, SPINNER, REFRESH
from gentab import SEED

from pathlib import Path
import json
import os
import copy

import optuna


class Tuner:
    def __init__(
        self,
        evaluator: Evaluator,
        n_trials: int,
        min_epochs: int = 300,
        max_epochs: int = 800,
        min_batch: int = 512,
        max_batch: int = 4096,
        max_tries_per_batch: int = 8192,
        timeout: int = None,
    ) -> None:
        self.seed = SEED
        self.evaluator = evaluator
        self.generator = evaluator.generator
        self.dataset = evaluator.generator.dataset
        self.n_trials = n_trials
        self.min_epochs = min_epochs
        self.max_epochs = max_epochs
        self.batch_sizes = [
            2**i
            for i in range(
                int(min_batch).bit_length() - 1, int(max_batch).bit_length() + 1
            )
            if 2**i >= min_batch and 2**i <= max_batch
        ]
        self.max_tries_per_batch = max_tries_per_batch
        self.folder = "tuning"
        self.timeout = timeout

    def __str__(self) -> str:
        return self.__class__.__name__

    def store_data(self, trial: optuna.trial.Trial) -> None:
        trial.set_user_attr("timing", self.generator.timer.history)
        trial.set_user_attr("dataset", copy.deepcopy(self.dataset))

    def objective(self, trial: optuna.trial.Trial) -> float:
        pass

    def save_to_disk(self):
        self.dataset.save_to_disk(self.generator, self.evaluator)

    def load_from_disk(self):
        self.dataset.load_from_disk(self.generator, self.evaluator)

    def tune(self) -> None:
        # pruner: optuna.pruners.BasePruner(optuna.pruners.NopPruner())

        self.study = optuna.create_study(
            study_name=self,
            direction="maximize",
            sampler=optuna.samplers.TPESampler(seed=self.seed),
        )

        self.study.optimize(
            self.objective, n_trials=self.n_trials, timeout=self.timeout
        )

        console.print("Number of finished trials: {}".format(len(self.study.trials)))

        console.print("Best trial:")
        self.trial = self.study.best_trial
        console.print("  Value: {}".format(self.trial.value))
        console.print("  Params: ")
        for key, value in self.trial.params.items():
            console.print("    {}: {}".format(key, value))

        # Add timing information
        self.trial.params["train_time"] = self.trial.user_attrs["timing"][0]
        self.trial.params["gen_time"] = self.trial.user_attrs["timing"][1]

        # Save generator parameters to JSON
        Path(self.folder).mkdir(parents=True, exist_ok=True)
        path = os.path.join(
            self.folder,
            str(self.dataset).lower()
            + "_"
            + str(self.generator).lower()
            + "_"
            + str(self.evaluator).lower()
            + ".json",
        )
        with open(path, "w") as fp:
            json.dump(self.trial.params, fp, indent=4)

        # Get best dataset object
        self.dataset = self.trial.user_attrs["dataset"]

    def get_tuning_info(self):
        # Load data from JSON
        path = os.path.join(
            self.folder,
            str(self.dataset).lower()
            + "_"
            + str(self.generator).lower()
            + "_"
            + str(self.evaluator).lower()
            + ".json",
        )
        with open(path, "r") as fp:
            return json.load(fp)



import torch
import numpy as np
from torch.utils.data import DataLoader
from .network import AutoEncoder, SwapNoiseCorrupter
from .data import SingleDataset
from .util import AverageMeter, EarlyStopping


def _init_dataloaders(data, datatype_info, batch_size, validation_ratio):
    """Split data into training set and validation set, construct dataloader for each."""
    n = len(data)
    cutoff = int(n * (1 - validation_ratio))

    train_ds = SingleDataset(data[:cutoff, :], datatype_info)
    valid_ds = SingleDataset(data[cutoff:, :], datatype_info)

    train_dl = DataLoader(
        train_ds, batch_size=batch_size, shuffle=True, pin_memory=True, drop_last=True
    )
    valid_dl = DataLoader(
        valid_ds, batch_size=batch_size, shuffle=True, pin_memory=True, drop_last=True
    )
    return train_dl, valid_dl


def _init_swap_noise_makers(datatype_info, swap_noise_probas):
    """Create swap noise corrupter for each input data type."""
    n_bins, n_cats, n_nums = (
        datatype_info["n_bins"],
        datatype_info["n_cats"],
        datatype_info["n_nums"],
    )
    if isinstance(swap_noise_probas, (float, int)):
        swap_noise_probas = [swap_noise_probas] * sum([n_bins, n_cats, n_nums])

    noise_makers = dict()
    if n_bins:
        noise_makers["bins"] = SwapNoiseCorrupter(swap_noise_probas[:n_bins])
    if n_cats:
        noise_makers["cats"] = SwapNoiseCorrupter(
            swap_noise_probas[n_bins : n_bins + n_cats]
        )
    if n_nums:
        noise_makers["nums"] = SwapNoiseCorrupter(swap_noise_probas[-n_nums:])
    return noise_makers


def _apply_noise(batch_data, noise_makers):
    """Apply swap noise on data."""
    noisy_batch, masks = dict(), dict()
    for typ in ["bins", "cats", "nums"]:
        if typ in batch_data:
            noisy_data, mask = noise_makers[typ](batch_data[typ])
            noisy_batch[typ] = noisy_data
            masks[typ] = mask
    return noisy_batch, masks


def _init_loss_weights(loss_weights, datatype_info, mask_loss_weight=2):
    if isinstance(loss_weights, dict):
        return loss_weights
    n_bins, n_cats, n_nums = (
        datatype_info["n_bins"],
        datatype_info["n_cats"],
        datatype_info["n_nums"],
    )
    loss_weights = dict()
    total = sum([n_bins, n_cats, n_nums])
    if n_bins:
        loss_weights["bins"] = n_bins / total
    if n_cats:
        loss_weights["cats"] = n_cats / total
    if n_nums:
        loss_weights["nums"] = n_nums / total
    loss_weights["mask"] = mask_loss_weight
    return loss_weights


def train(
    network_cfg_or_network,
    data,
    datatype_info,
    swap_noise_probas,
    validation_ratio,
    batch_size=128,
    max_epochs=1024,
    early_stopping_rounds=100,
    eval_verbose=10,
    verbose=2,
    optimizer_fn=torch.optim.Adam,
    optimizer_params={"lr": 3e-4},
    scheduler_fn=torch.optim.lr_scheduler.ReduceLROnPlateau,
    scheduler_params=dict(),
    loss_weights=None,
    mask_loss_weight=2,
    device="cpu",
    model_checkpoint="./model_checkpoint.pth",
):
    # make sure network is ready and on device.
    assert isinstance(
        network_cfg_or_network, (dict, AutoEncoder)
    ), "either supply network itself, or a recepit to make one."
    if isinstance(network_cfg_or_network, dict):
        network = AutoEncoder(**network_cfg_or_network).to(device)
    else:
        network = network_cfg_or_network.to(device)

    # prepare to train
    train_dl, valid_dl = _init_dataloaders(
        data, datatype_info, batch_size, validation_ratio
    )
    noise_makers = _init_swap_noise_makers(datatype_info, swap_noise_probas)
    loss_weights = _init_loss_weights(loss_weights, datatype_info, mask_loss_weight)
    optimizer = optimizer_fn(network.parameters(), **optimizer_params)
    scheduler = (
        None
        if not scheduler_fn
        else scheduler_fn(optimizer, verbose=verbose, **scheduler_params)
    )
    earlystop = (
        None
        if early_stopping_rounds == 0
        else EarlyStopping(patience=early_stopping_rounds, verbose=verbose)
    )

    network = network.to(device)

    best_score = float("inf")
    # training network
    for epoch in range(max_epochs):

        # train step
        network.train()
        meter = AverageMeter()
        for i, x in enumerate(train_dl):
            for k in x:
                x[k] = x[k].to(device, non_blocking=True)
            noisy_x, masks = _apply_noise(x, noise_makers)
            optimizer.zero_grad()
            reconstruction, predicted_mask = network(noisy_x)
            loss = network.loss(x, masks, reconstruction, predicted_mask, loss_weights)
            loss.backward()
            optimizer.step()
            meter.update(loss.item())
            if verbose > 1:
                print(
                    "\repoch {:4d} - batch {:4d} - train loss {:6.4f}".format(
                        epoch, i, meter.avg
                    ),
                    end="",
                )
        train_loss = meter.overall_avg

        # validation step
        meter.reset()
        with torch.no_grad():
            for i, x in enumerate(valid_dl):
                for k in x:
                    x[k] = x[k].to(device, non_blocking=True)
                noisy_x, masks = _apply_noise(x, noise_makers)
                reconstruction, predicted_mask = network(noisy_x)
                loss = network.loss(
                    x, masks, reconstruction, predicted_mask, loss_weights
                )
                meter.update(loss.item())
                if verbose > 1:
                    print(
                        "\repoch {:4d} - batch {:4d} - valid loss {:6.4f}".format(
                            epoch, i, meter.avg
                        ),
                        end="",
                    )
        valid_loss = meter.overall_avg

        if verbose and epoch % eval_verbose == 0:
            print(
                "\repoch {:4d} - train loss {:6.4f} - valid loss {:6.4f}".format(
                    epoch, train_loss, valid_loss
                )
            )

        # adjust learning rate if neccessary
        if scheduler is not None:
            if isinstance(scheduler, torch.optim.lr_scheduler.ReduceLROnPlateau):
                scheduler.step(valid_loss)
            else:
                scheduler.step(epoch)

        # checkpointing
        if valid_loss < best_score:
            best_score = valid_loss
            torch.save({"model": network.state_dict()}, model_checkpoint)

        # early stopping
        if earlystop.step(valid_loss):
            break

    # retrieve the best weights
    model_state_dict = torch.load(model_checkpoint)
    network.load_state_dict(model_state_dict["model"])
    return network


def featurize(network, data, datatype_info, batch_size, device="cpu"):
    ds = SingleDataset(data, datatype_info)
    dl = DataLoader(
        ds, batch_size=batch_size, shuffle=False, pin_memory=True, drop_last=False
    )
    features = []
    with torch.no_grad():
        for i, x in enumerate(dl):
            for k in x:
                x[k] = x[k].to(device, non_blocking=True)
            batch_featurs = network.featurize(x)
            features.append(batch_featurs.detach().cpu().numpy())
    features = np.vstack(features)
    return features



import typing as tp

import numpy as np
import pandas as pd
import torch

from transformers import AutoTokenizer
from transformers import logging

logging.set_verbosity_error()


def _array_to_dataframe(
    data: tp.Union[pd.DataFrame, np.ndarray], columns=None
) -> pd.DataFrame:
    """Converts a Numpy Array to a Pandas DataFrame

    Args:
        data: Pandas DataFrame or Numpy NDArray
        columns: If data is a Numpy Array, columns needs to be a list of all column names

    Returns:
        Pandas DataFrame with the given data
    """
    if isinstance(data, pd.DataFrame):
        return data

    assert isinstance(
        data, np.ndarray
    ), "Input needs to be a Pandas DataFrame or a Numpy NDArray"
    assert (
        columns
    ), "To convert the data into a Pandas DataFrame, a list of column names has to be given!"
    assert len(columns) == len(
        data[0]
    ), "%d column names are given, but array has %d columns!" % (
        len(columns),
        len(data[0]),
    )

    return pd.DataFrame(data=data, columns=columns)


def _get_column_distribution(df: pd.DataFrame, col: str) -> tp.Union[list, dict]:
    """Returns the distribution of a given column. If continuous, returns a list of all values.
        If categorical, returns a dictionary in form {"A": 0.6, "B": 0.4}

    Args:
        df: pandas DataFrame
        col: name of the column

    Returns:
        Distribution of the column
    """
    if df[col].dtype == "float":
        col_dist = df[col].to_list()
    else:
        col_dist = df[col].value_counts(1).to_dict()
    return col_dist


def _convert_tokens_to_text(
    tokens: tp.List[torch.Tensor], tokenizer: AutoTokenizer
) -> tp.List[str]:
    """Decodes the tokens back to strings

    Args:
        tokens: List of tokens to decode
        tokenizer: Tokenizer used for decoding

    Returns:
        List of decoded strings
    """
    # Convert tokens to text
    text_data = [tokenizer.decode(t) for t in tokens]

    # Clean text
    text_data = [d.replace("<|endoftext|>", "") for d in text_data]
    text_data = [d.replace("\n", " ") for d in text_data]
    text_data = [d.replace("\r", "") for d in text_data]

    return text_data


def _convert_text_to_tabular_data(
    text: tp.List[str], df_gen: pd.DataFrame
) -> pd.DataFrame:
    """Converts the sentences back to tabular data

    Args:
        text: List of the tabular data in text form
        df_gen: Pandas DataFrame where the tabular data is appended

    Returns:
        Pandas DataFrame with the tabular data from the text appended
    """
    columns = df_gen.columns.to_list()
    result_list = []
    # Convert text to tabular data
    for t in text:
        features = t.split(",")
        td = dict.fromkeys(columns)

        # Transform all features back to tabular data
        for f in features:
            values = f.strip().split(" ")
            if values[0] in columns and not td[values[0]]:
                try:
                    td[values[0]] = [values[1]]
                except IndexError:
                    # print("An Index Error occurred - if this happends a lot, consider fine-tuning your model further.")
                    pass
        result_list.append(pd.DataFrame(td))
        # df_gen = pd.concat([df_gen, pd.DataFrame(td)], ignore_index=True, axis=0)
    generated_df = pd.concat(result_list, ignore_index=True, axis=0)
    df_gen = pd.concat([df_gen, generated_df], ignore_index=True, axis=0)
    return df_gen



from . import Generator
from gentab.utils import console, DEVICE

from be_great import GReaT as GR
import pandas as pd
import typing as tp


class GReaT(Generator):
    """GReaT Class

    The GReaT class handles the whole generation flow. It is used to fine-tune a large language model for tabular data,
    and to sample synthetic tabular data.

    Attributes:
        llm (str): HuggingFace checkpoint of a pretrained large language model, used a basis of our model
        tokenizer (AutoTokenizer): Tokenizer, automatically downloaded from llm-checkpoint
        model (AutoModelForCausalLM): Large language model, automatically downloaded from llm-checkpoint
        experiment_dir (str): Directory, where the training checkpoints will be saved
        epochs (int): Number of epochs to fine-tune the model
        batch_size (int): Batch size used for fine-tuning
        train_hyperparameters (dict): Additional hyperparameters added to the TrainingArguments used by the
         HuggingFaceLibrary, see here the full list of all possible values
         https://huggingface.co/docs/transformers/main/en/main_classes/trainer#transformers.TrainingArguments
        columns (list): List of all features/columns of the tabular dataset
        num_cols (list): List of all numerical features/columns of the tabular dataset
        conditional_col (str): Name of a feature/column on which the sampling can be conditioned
        conditional_col_dist (dict | list): Distribution of the feature/column specified by condtional_col
    """

    def __init__(
        self,
        dataset,
        llm: str = "distilgpt2",
        experiment_dir: str = "trainer_great",
        epochs: int = 15,
        batch_size: int = 32,
        max_tries_per_batch: int = 4096,
        efficient_finetuning: str = "",
        start_col: tp.Optional[str] = "",
        start_col_dist: tp.Optional[tp.Union[dict, list]] = None,
        temperature: float = 0.6,
        k: int = 100,
        max_length: int = 1024,
        drop_nan: bool = False,
        device: str = DEVICE,
        n_samples: int = 8192,
    ) -> None:
        super().__init__(dataset, batch_size, max_tries_per_batch)
        self.data = self.dataset.get_single_df()
        self.llm = llm
        self.experiment_dir = experiment_dir
        self.epochs = epochs
        self.efficient_finetuning = efficient_finetuning
        self.start_col = start_col
        self.start_col_dist = start_col_dist
        self.temperature = temperature
        self.k = k
        self.drop_nan = drop_nan
        self.max_length = max_length
        self.device = device
        self.n_samples = n_samples

        self.model = GR(
            llm=self.llm,
            experiment_dir=self.experiment_dir,
            batch_size=self.batch_size,
            epochs=self.epochs,
            efficient_finetuning=self.efficient_finetuning,
        )

    def preprocess(self) -> None:
        return super().preprocess()

    def train(self) -> None:
        self.model.fit(data=self.data)

    def sample(self) -> pd.DataFrame:
        return self.model.sample(
            n_samples=self.n_samples,
            start_col=self.start_col,
            start_col_dist=self.start_col_dist,
            temperature=self.temperature,
            k=self.k,
            max_length=self.max_length,
            device=self.device,
            drop_nan=self.drop_nan,
        )



import random
import typing as tp

from datasets import Dataset
from dataclasses import dataclass
from transformers import DataCollatorWithPadding
from transformers import logging

logging.set_verbosity_error()


class TabulaDataset(Dataset):
    """Tabula Dataset

    The TabulaDataset overwrites the _getitem function of the HuggingFace Dataset Class to include the permutation step.

    Attributes:
        tokenizer (AutoTokenizer): Tokenizer from HuggingFace
    """

    def set_tokenizer(self, tokenizer):
        """Set the Tokenizer

        Args:
            tokenizer: Tokenizer from HuggingFace
        """
        self.tokenizer = tokenizer

    def _getitem(
        self, key: tp.Union[int, slice, str], decoded: bool = True, **kwargs
    ) -> tp.Union[tp.Dict, tp.List]:
        """Get Item from Tabular Data

        Get one instance of the tabular data, permuted, converted to text and tokenized.
        """
        # If int, what else?
        row = self._data.fast_slice(key, 1)

        shuffle_idx = list(range(row.num_columns))
        random.shuffle(shuffle_idx)

        shuffled_text = ", ".join([
            "%s %s" % (row.column_names[i], str(row.columns[i].to_pylist()[0]).strip())
            for i in shuffle_idx
        ])

        tokenized_text = self.tokenizer(shuffled_text)
        return tokenized_text

    def __getitems__(self, keys: tp.Union[int, slice, str, list]):
        if isinstance(keys, list):
            return [self._getitem(key) for key in keys]
        else:
            return self._getitem(keys)


@dataclass
class TabulaDataCollator(DataCollatorWithPadding):
    """Tabula Data Collator

    Overwrites the DataCollatorWithPadding to also pad the labels and not only the input_ids
    """

    def __call__(self, features: tp.List[tp.Dict[str, tp.Any]]):
        batch = self.tokenizer.pad(
            features,
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors=self.return_tensors,
        )
        batch["labels"] = batch["input_ids"].clone()
        # Use following print to observe encoded token sequence length
        # print("labels shape: ", batch["labels"].shape)
        return batch



import numpy as np
import pandas as pd
import torch
import torch.utils.data
import torch.optim as optim
from torch.optim import Adam
from torch.nn import functional as F
from torch.nn import (
    Dropout,
    LeakyReLU,
    Linear,
    Module,
    ReLU,
    Sequential,
    Conv2d,
    ConvTranspose2d,
    Sigmoid,
    init,
    BCELoss,
    CrossEntropyLoss,
    SmoothL1Loss,
    LayerNorm,
)
from .transformer import ImageTransformer, DataTransformer
from gentab.generators.ctabgplus.privacy_utils.rdp_accountant import (
    compute_rdp,
    get_privacy_spent,
)


class Classifier(Module):
    def __init__(self, input_dim, dis_dims, st_ed):
        super(Classifier, self).__init__()
        dim = input_dim - (st_ed[1] - st_ed[0])
        seq = []
        self.str_end = st_ed
        for item in list(dis_dims):
            seq += [Linear(dim, item), LeakyReLU(0.2), Dropout(0.5)]
            dim = item

        if (st_ed[1] - st_ed[0]) == 1:
            seq += [Linear(dim, 1)]

        elif (st_ed[1] - st_ed[0]) == 2:
            seq += [Linear(dim, 1), Sigmoid()]
        else:
            seq += [Linear(dim, (st_ed[1] - st_ed[0]))]

        self.seq = Sequential(*seq)

    def forward(self, input):

        label = None

        if (self.str_end[1] - self.str_end[0]) == 1:
            label = input[:, self.str_end[0] : self.str_end[1]]
        else:
            label = torch.argmax(input[:, self.str_end[0] : self.str_end[1]], axis=-1)

        new_imp = torch.cat(
            (input[:, : self.str_end[0]], input[:, self.str_end[1] :]), 1
        )

        if ((self.str_end[1] - self.str_end[0]) == 2) | (
            (self.str_end[1] - self.str_end[0]) == 1
        ):
            return self.seq(new_imp).view(-1), label
        else:
            return self.seq(new_imp), label


def apply_activate(data, output_info):
    data_t = []
    st = 0
    for item in output_info:
        if item[1] == "tanh":
            ed = st + item[0]
            data_t.append(torch.tanh(data[:, st:ed]))
            st = ed
        elif item[1] == "softmax":
            ed = st + item[0]
            data_t.append(F.gumbel_softmax(data[:, st:ed], tau=0.2))
            st = ed
    return torch.cat(data_t, dim=1)


def get_st_ed(target_col_index, output_info):
    st = 0
    c = 0
    tc = 0

    for item in output_info:
        if c == target_col_index:
            break
        if item[1] == "tanh":
            st += item[0]
            if item[2] == "yes_g":
                c += 1
        elif item[1] == "softmax":
            st += item[0]
            c += 1
        tc += 1

    ed = st + output_info[tc][0]

    return (st, ed)


def random_choice_prob_index_sampling(probs, col_idx):
    option_list = []
    for i in col_idx:
        pp = probs[i]
        option_list.append(np.random.choice(np.arange(len(probs[i])), p=pp))

    return np.array(option_list).reshape(col_idx.shape)


def random_choice_prob_index(a, axis=1):
    r = np.expand_dims(np.random.rand(a.shape[1 - axis]), axis=axis)
    return (a.cumsum(axis=axis) > r).argmax(axis=axis)


def maximum_interval(output_info):
    max_interval = 0
    for item in output_info:
        max_interval = max(max_interval, item[0])
    return max_interval


class Cond(object):
    def __init__(self, data, output_info):

        self.model = []
        st = 0
        counter = 0
        for item in output_info:

            if item[1] == "tanh":
                st += item[0]
                continue
            elif item[1] == "softmax":
                ed = st + item[0]
                counter += 1
                self.model.append(np.argmax(data[:, st:ed], axis=-1))
                st = ed

        self.interval = []
        self.n_col = 0
        self.n_opt = 0
        st = 0
        self.p = np.zeros((counter, maximum_interval(output_info)))
        self.p_sampling = []
        for item in output_info:
            if item[1] == "tanh":
                st += item[0]
                continue
            elif item[1] == "softmax":
                ed = st + item[0]
                tmp = np.sum(data[:, st:ed], axis=0)
                tmp_sampling = np.sum(data[:, st:ed], axis=0)
                tmp = np.log(tmp + 1)
                tmp = tmp / np.sum(tmp)
                tmp_sampling = tmp_sampling / np.sum(tmp_sampling)
                self.p_sampling.append(tmp_sampling)
                self.p[self.n_col, : item[0]] = tmp
                self.interval.append((self.n_opt, item[0]))
                self.n_opt += item[0]
                self.n_col += 1
                st = ed

        self.interval = np.asarray(self.interval)

    def sample_train(self, batch):
        if self.n_col == 0:
            return None
        batch = batch

        idx = np.random.choice(np.arange(self.n_col), batch)

        vec = np.zeros((batch, self.n_opt), dtype="float32")
        mask = np.zeros((batch, self.n_col), dtype="float32")
        mask[np.arange(batch), idx] = 1
        opt1prime = random_choice_prob_index(self.p[idx])
        for i in np.arange(batch):
            vec[i, self.interval[idx[i], 0] + opt1prime[i]] = 1

        return vec, mask, idx, opt1prime

    def sample(self, batch):
        if self.n_col == 0:
            return None
        batch = batch

        idx = np.random.choice(np.arange(self.n_col), batch)

        vec = np.zeros((batch, self.n_opt), dtype="float32")
        opt1prime = random_choice_prob_index_sampling(self.p_sampling, idx)

        for i in np.arange(batch):
            vec[i, self.interval[idx[i], 0] + opt1prime[i]] = 1

        return vec


def cond_loss(data, output_info, c, m):
    loss = []
    st = 0
    st_c = 0
    for item in output_info:
        if item[1] == "tanh":
            st += item[0]
            continue

        elif item[1] == "softmax":
            ed = st + item[0]
            ed_c = st_c + item[0]
            tmp = F.cross_entropy(
                data[:, st:ed], torch.argmax(c[:, st_c:ed_c], dim=1), reduction="none"
            )
            loss.append(tmp)
            st = ed
            st_c = ed_c

    loss = torch.stack(loss, dim=1)
    return (loss * m).sum() / data.size()[0]


class Sampler(object):
    def __init__(self, data, output_info):
        super(Sampler, self).__init__()
        self.data = data
        self.model = []
        self.n = len(data)
        st = 0
        for item in output_info:
            if item[1] == "tanh":
                st += item[0]
                continue
            elif item[1] == "softmax":
                ed = st + item[0]
                tmp = []
                for j in range(item[0]):
                    tmp.append(np.nonzero(data[:, st + j])[0])
                self.model.append(tmp)
                st = ed

    def sample(self, n, col, opt):
        if col is None:
            idx = np.random.choice(np.arange(self.n), n)
            return self.data[idx]
        idx = []
        for c, o in zip(col, opt):
            idx.append(np.random.choice(self.model[c][o]))
        return self.data[idx]


class Discriminator(Module):
    def __init__(self, side, layers):
        super(Discriminator, self).__init__()
        self.side = side
        info = len(layers) - 2
        self.seq = Sequential(*layers)
        self.seq_info = Sequential(*layers[:info])

    def forward(self, input):
        return (self.seq(input)), self.seq_info(input)


class Generator(Module):
    def __init__(self, side, layers):
        super(Generator, self).__init__()
        self.side = side
        self.seq = Sequential(*layers)

    def forward(self, input_):
        return self.seq(input_)


def determine_layers_disc(side, num_channels):
    assert side >= 4 and side <= 64

    layer_dims = [(1, side), (num_channels, side // 2)]

    while layer_dims[-1][1] > 3 and len(layer_dims) < 4:
        layer_dims.append((layer_dims[-1][0] * 2, layer_dims[-1][1] // 2))

    layerNorms = []
    num_c = num_channels
    num_s = side / 2
    for l in range(len(layer_dims) - 1):
        layerNorms.append([int(num_c), int(num_s), int(num_s)])
        num_c = num_c * 2
        num_s = num_s / 2

    layers_D = []

    for prev, curr, ln in zip(layer_dims, layer_dims[1:], layerNorms):
        layers_D += [
            Conv2d(prev[0], curr[0], 4, 2, 1, bias=False),
            LayerNorm(ln),
            LeakyReLU(0.2, inplace=True),
        ]

    layers_D += [Conv2d(layer_dims[-1][0], 1, layer_dims[-1][1], 1, 0), ReLU(True)]

    return layers_D


def determine_layers_gen(side, random_dim, num_channels):
    assert side >= 4 and side <= 64

    layer_dims = [(1, side), (num_channels, side // 2)]

    while layer_dims[-1][1] > 3 and len(layer_dims) < 4:
        layer_dims.append((layer_dims[-1][0] * 2, layer_dims[-1][1] // 2))

    layerNorms = []

    num_c = num_channels * (2 ** (len(layer_dims) - 2))
    num_s = int(side / (2 ** (len(layer_dims) - 1)))
    for l in range(len(layer_dims) - 1):
        layerNorms.append([int(num_c), int(num_s), int(num_s)])
        num_c = num_c / 2
        num_s = num_s * 2

    layers_G = [
        ConvTranspose2d(
            random_dim,
            layer_dims[-1][0],
            layer_dims[-1][1],
            1,
            0,
            output_padding=0,
            bias=False,
        )
    ]

    for prev, curr, ln in zip(
        reversed(layer_dims), reversed(layer_dims[:-1]), layerNorms
    ):
        layers_G += [
            LayerNorm(ln),
            ReLU(True),
            ConvTranspose2d(prev[0], curr[0], 4, 2, 1, output_padding=0, bias=True),
        ]
    return layers_G


def slerp(val, low, high):
    low_norm = low / torch.norm(low, dim=1, keepdim=True)
    high_norm = high / torch.norm(high, dim=1, keepdim=True)
    omega = torch.acos((low_norm * high_norm).sum(1)).view(val.size(0), 1)
    so = torch.sin(omega)
    res = (torch.sin((1.0 - val) * omega) / so) * low + (
        torch.sin(val * omega) / so
    ) * high

    return res


def calc_gradient_penalty_slerp(
    netD, real_data, fake_data, transformer, device="cpu", lambda_=10
):
    batchsize = real_data.shape[0]
    alpha = torch.rand(batchsize, 1, device=device)
    interpolates = slerp(alpha, real_data, fake_data)
    interpolates = interpolates.to(device)
    interpolates = transformer.transform(interpolates)
    interpolates = torch.autograd.Variable(interpolates, requires_grad=True)
    disc_interpolates, _ = netD(interpolates)

    gradients = torch.autograd.grad(
        outputs=disc_interpolates,
        inputs=interpolates,
        grad_outputs=torch.ones(disc_interpolates.size()).to(device),
        create_graph=True,
        retain_graph=True,
        only_inputs=True,
    )[0]

    gradients_norm = gradients.norm(2, dim=1)
    gradient_penalty = ((gradients_norm - 1) ** 2).mean() * lambda_

    return gradient_penalty


def weights_init(m):
    classname = m.__class__.__name__

    if classname.find("Conv") != -1:
        init.normal_(m.weight.data, 0.0, 0.02)

    elif classname.find("BatchNorm") != -1:
        init.normal_(m.weight.data, 1.0, 0.02)
        init.constant_(m.bias.data, 0)


class CTABGANSynthesizer:
    def __init__(
        self,
        class_dim=(256, 256, 256, 256),
        random_dim=100,
        num_channels=64,
        l2scale=1e-5,
        batch_size=500,
        epochs=150,
    ):

        self.random_dim = random_dim
        self.class_dim = class_dim
        self.num_channels = num_channels
        self.dside = None
        self.gside = None
        self.l2scale = l2scale
        self.batch_size = batch_size
        self.epochs = epochs
        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

    def fit(
        self,
        train_data=pd.DataFrame,
        categorical=[],
        mixed={},
        general=[],
        non_categorical=[],
        type={},
        progress=None,
        task=None,
    ):

        problem_type = None
        target_index = None
        if type:
            problem_type = list(type.keys())[0]
            if problem_type:
                target_index = train_data.columns.get_loc(type[problem_type])

        self.transformer = DataTransformer(
            train_data=train_data,
            categorical_list=categorical,
            mixed_dict=mixed,
            general_list=general,
            non_categorical_list=non_categorical,
        )
        self.transformer.fit()
        train_data = self.transformer.transform(train_data.values)
        data_sampler = Sampler(train_data, self.transformer.output_info)
        data_dim = self.transformer.output_dim
        self.cond_generator = Cond(train_data, self.transformer.output_info)

        sides = [4, 8, 16, 24, 32, 64, 128]
        col_size_d = data_dim + self.cond_generator.n_opt
        for i in sides:
            if i * i >= col_size_d:
                self.dside = i
                break

        sides = [4, 8, 16, 24, 32, 64, 128]
        col_size_g = data_dim
        for i in sides:
            if i * i >= col_size_g:
                self.gside = i
                break

        layers_G = determine_layers_gen(
            self.gside, self.random_dim + self.cond_generator.n_opt, self.num_channels
        )
        layers_D = determine_layers_disc(self.dside, self.num_channels)

        self.generator = Generator(self.gside, layers_G).to(self.device)
        discriminator = Discriminator(self.dside, layers_D).to(self.device)
        optimizer_params = dict(
            lr=2e-4, betas=(0.5, 0.9), eps=1e-3, weight_decay=self.l2scale
        )
        optimizerG = Adam(self.generator.parameters(), **optimizer_params)
        optimizerD = Adam(discriminator.parameters(), **optimizer_params)

        st_ed = None
        classifier = None
        optimizerC = None
        if target_index != None:
            st_ed = get_st_ed(target_index, self.transformer.output_info)
            classifier = Classifier(data_dim, self.class_dim, st_ed).to(self.device)
            optimizerC = optim.Adam(classifier.parameters(), **optimizer_params)

        self.generator.apply(weights_init)
        discriminator.apply(weights_init)

        self.Gtransformer = ImageTransformer(self.gside)
        self.Dtransformer = ImageTransformer(self.dside)

        epsilon = 0
        epoch = 0
        steps = 0
        ci = 1

        steps_per_epoch = max(1, len(train_data) // self.batch_size)
        for i in range(self.epochs):
            for id_ in range(steps_per_epoch):

                for _ in range(ci):
                    noisez = torch.randn(
                        self.batch_size, self.random_dim, device=self.device
                    )
                    condvec = self.cond_generator.sample_train(self.batch_size)

                    c, m, col, opt = condvec
                    c = torch.from_numpy(c).to(self.device)
                    m = torch.from_numpy(m).to(self.device)
                    noisez = torch.cat([noisez, c], dim=1)
                    noisez = noisez.view(
                        self.batch_size,
                        self.random_dim + self.cond_generator.n_opt,
                        1,
                        1,
                    )

                    perm = np.arange(self.batch_size)
                    np.random.shuffle(perm)
                    real = data_sampler.sample(self.batch_size, col[perm], opt[perm])
                    c_perm = c[perm]

                    real = torch.from_numpy(real.astype("float32")).to(self.device)

                    fake = self.generator(noisez)
                    faket = self.Gtransformer.inverse_transform(fake)
                    fakeact = apply_activate(faket, self.transformer.output_info)

                    fake_cat = torch.cat([fakeact, c], dim=1)
                    real_cat = torch.cat([real, c_perm], dim=1)

                    real_cat_d = self.Dtransformer.transform(real_cat)
                    fake_cat_d = self.Dtransformer.transform(fake_cat)

                    optimizerD.zero_grad()

                    d_real, _ = discriminator(real_cat_d)

                    d_real = -torch.mean(d_real)
                    d_real.backward()

                    d_fake, _ = discriminator(fake_cat_d)

                    d_fake = torch.mean(d_fake)

                    d_fake.backward()

                    pen = calc_gradient_penalty_slerp(
                        discriminator,
                        real_cat,
                        fake_cat,
                        self.Dtransformer,
                        self.device,
                    )

                    pen.backward()

                    optimizerD.step()

                noisez = torch.randn(
                    self.batch_size, self.random_dim, device=self.device
                )

                condvec = self.cond_generator.sample_train(self.batch_size)

                c, m, col, opt = condvec
                c = torch.from_numpy(c).to(self.device)
                m = torch.from_numpy(m).to(self.device)
                noisez = torch.cat([noisez, c], dim=1)
                noisez = noisez.view(
                    self.batch_size, self.random_dim + self.cond_generator.n_opt, 1, 1
                )

                optimizerG.zero_grad()

                fake = self.generator(noisez)
                faket = self.Gtransformer.inverse_transform(fake)
                fakeact = apply_activate(faket, self.transformer.output_info)

                fake_cat = torch.cat([fakeact, c], dim=1)
                fake_cat = self.Dtransformer.transform(fake_cat)

                y_fake, info_fake = discriminator(fake_cat)

                cross_entropy = cond_loss(faket, self.transformer.output_info, c, m)

                _, info_real = discriminator(real_cat_d)

                g = -torch.mean(y_fake) + cross_entropy
                g.backward(retain_graph=True)
                loss_mean = torch.norm(
                    torch.mean(info_fake.view(self.batch_size, -1), dim=0)
                    - torch.mean(info_real.view(self.batch_size, -1), dim=0),
                    1,
                )
                loss_std = torch.norm(
                    torch.std(info_fake.view(self.batch_size, -1), dim=0)
                    - torch.std(info_real.view(self.batch_size, -1), dim=0),
                    1,
                )
                loss_info = loss_mean + loss_std
                loss_info.backward()
                optimizerG.step()

                if problem_type:

                    fake = self.generator(noisez)

                    faket = self.Gtransformer.inverse_transform(fake)

                    fakeact = apply_activate(faket, self.transformer.output_info)

                    real_pre, real_label = classifier(real)
                    fake_pre, fake_label = classifier(fakeact)

                    c_loss = CrossEntropyLoss()

                    if (st_ed[1] - st_ed[0]) == 1:
                        c_loss = SmoothL1Loss()
                        real_label = real_label.type_as(real_pre)
                        fake_label = fake_label.type_as(fake_pre)
                        real_label = torch.reshape(real_label, real_pre.size())
                        fake_label = torch.reshape(fake_label, fake_pre.size())

                    elif (st_ed[1] - st_ed[0]) == 2:
                        c_loss = BCELoss()
                        real_label = real_label.type_as(real_pre)
                        fake_label = fake_label.type_as(fake_pre)

                    loss_cc = c_loss(real_pre, real_label)
                    loss_cg = c_loss(fake_pre, fake_label)

                    optimizerG.zero_grad()
                    loss_cg.backward()
                    optimizerG.step()

                    optimizerC.zero_grad()
                    loss_cc.backward()
                    optimizerC.step()

            epoch += 1
            progress.update(task, completed=i)

    def sample(self, n):

        self.generator.eval()

        output_info = self.transformer.output_info
        steps = n // self.batch_size + 1

        data = []

        for i in range(steps):
            noisez = torch.randn(self.batch_size, self.random_dim, device=self.device)
            condvec = self.cond_generator.sample(self.batch_size)
            c = condvec
            c = torch.from_numpy(c).to(self.device)
            noisez = torch.cat([noisez, c], dim=1)
            noisez = noisez.view(
                self.batch_size, self.random_dim + self.cond_generator.n_opt, 1, 1
            )

            fake = self.generator(noisez)
            faket = self.Gtransformer.inverse_transform(fake)
            fakeact = apply_activate(faket, output_info)
            data.append(fakeact.detach().cpu().numpy())

        data = np.concatenate(data, axis=0)
        result, resample = self.transformer.inverse_transform(data)

        while len(result) < n:
            data_resample = []
            steps_left = resample // self.batch_size + 1

            for i in range(steps_left):
                noisez = torch.randn(
                    self.batch_size, self.random_dim, device=self.device
                )
                condvec = self.cond_generator.sample(self.batch_size)
                c = condvec
                c = torch.from_numpy(c).to(self.device)
                noisez = torch.cat([noisez, c], dim=1)
                noisez = noisez.view(
                    self.batch_size, self.random_dim + self.cond_generator.n_opt, 1, 1
                )

                fake = self.generator(noisez)
                faket = self.Gtransformer.inverse_transform(fake)
                fakeact = apply_activate(faket, output_info)
                data_resample.append(fakeact.detach().cpu().numpy())

            data_resample = np.concatenate(data_resample, axis=0)

            res, resample = self.transformer.inverse_transform(data_resample)
            result = np.concatenate([result, res], axis=0)

        return result[0:n]



from gentab.generators import (
    ROS,
    SMOTE,
    ADASYN,
    TVAE,
    CTGAN,
    GaussianCopula,
    CopulaGAN,
    CTABGAN,
    CTABGANPlus,
    AutoDiffusion,
    ForestDiffusion,
    Tabula,
    GReaT,
)
from gentab.evaluators import KNN, LightGBM, XGBoost, MLP
from gentab.tuners import (
    SMOTETuner,
    ADASYNTuner,
    TVAETuner,
    CTGANTuner,
    GaussianCopulaTuner,
    CopulaGANTuner,
    CTABGANTuner,
    CTABGANPlusTuner,
    AutoDiffusionTuner,
    ForestDiffusionTuner,
    TabulaTuner,
    GReaTTuner,
)

from gentab.data import Config, Dataset
from gentab.utils import console

config = Config("configs/adult.json")

dataset = Dataset(config)
dataset.merge_classes({"<=50K": ["<=50K."], ">50K": [">50K."]})

trials = 10

console.print(dataset.class_counts(), dataset.row_count())
generator = SMOTE(dataset)
evaluator = XGBoost(generator)
tuner = SMOTETuner(evaluator, trials)
tuner.tune()
tuner.save_to_disk()
console.print(dataset.generated_class_counts(), dataset.generated_row_count())

console.print(dataset.class_counts(), dataset.row_count())
generator = ADASYN(dataset, sampling_strategy="minority")
evaluator = XGBoost(generator)
tuner = ADASYNTuner(evaluator, trials)
tuner.tune()
tuner.save_to_disk()
console.print(dataset.generated_class_counts(), dataset.generated_row_count())

console.print(dataset.class_counts(), dataset.row_count())
generator = TVAE(dataset)
evaluator = XGBoost(generator)
tuner = TVAETuner(evaluator, trials)
tuner.tune()
tuner.save_to_disk()
console.print(dataset.generated_class_counts(), dataset.generated_row_count())

console.print(dataset.class_counts(), dataset.row_count())
generator = CTGAN(dataset)
evaluator = XGBoost(generator)
tuner = CTGANTuner(evaluator, trials)
tuner.tune()
tuner.save_to_disk()
console.print(dataset.generated_class_counts(), dataset.generated_row_count())

console.print(dataset.class_counts(), dataset.row_count())
generator = GaussianCopula(dataset)
evaluator = XGBoost(generator)
tuner = GaussianCopulaTuner(evaluator, trials)
tuner.tune()
tuner.save_to_disk()
console.print(dataset.generated_class_counts(), dataset.generated_row_count())

console.print(dataset.class_counts(), dataset.row_count())
generator = CopulaGAN(dataset)
evaluator = XGBoost(generator)
tuner = CopulaGANTuner(evaluator, trials)
tuner.tune()
tuner.save_to_disk()
console.print(dataset.generated_class_counts(), dataset.generated_row_count())

console.print(dataset.class_counts(), dataset.row_count())
generator = CTABGAN(
    dataset,
    test_ratio=0.10,
)
evaluator = XGBoost(generator)
tuner = CTABGANTuner(evaluator, trials)
tuner.tune()
tuner.save_to_disk()
console.print(dataset.generated_class_counts(), dataset.generated_row_count())

console.print(dataset.class_counts(), dataset.row_count())
generator = CTABGANPlus(
    dataset,
    test_ratio=0.10,
)
evaluator = XGBoost(generator)
tuner = CTABGANPlusTuner(evaluator, trials)
tuner.tune()
tuner.save_to_disk()
console.print(dataset.generated_class_counts(), dataset.generated_row_count())

console.print(dataset.class_counts(), dataset.row_count())
generator = AutoDiffusion(dataset)
evaluator = XGBoost(generator)
tuner = AutoDiffusionTuner(evaluator, trials)
tuner.tune()
tuner.save_to_disk()
console.print(dataset.generated_class_counts(), dataset.generated_row_count())

console.print(dataset.class_counts(), dataset.row_count())
generator = ForestDiffusion(dataset, n_jobs=1, duplicate_K=4, n_estimators=100)
evaluator = XGBoost(generator)
tuner = ForestDiffusionTuner(evaluator, trials)
tuner.tune()
tuner.save_to_disk()
console.print(dataset.generated_class_counts(), dataset.generated_row_count())

console.print(dataset.class_counts(), dataset.row_count())
generator = GReaT(
    dataset,
    epochs=15,
    max_length=1024,
    temperature=0.6,
    batch_size=32,
    max_tries_per_batch=4096,
    n_samples=8192,
)
evaluator = XGBoost(generator)
tuner = GReaTTuner(evaluator, trials, min_epochs=15, max_epochs=30)
tuner.tune()
tuner.save_to_disk()
console.print(dataset.generated_class_counts(), dataset.generated_row_count())

console.print(dataset.class_counts(), dataset.row_count())
generator = Tabula(
    dataset,
    epochs=15,
    max_length=1024,
    temperature=0.6,
    batch_size=32,
    max_tries_per_batch=16384,
    n_samples=8192,
)
evaluator = XGBoost(generator)
tuner = TabulaTuner(
    evaluator, trials, min_epochs=15, max_epochs=30, max_tries_per_batch=16384
)
tuner.tune()
tuner.save_to_disk()
console.print(dataset.generated_class_counts(), dataset.generated_row_count())
